from __future__ import annotations

import glob
import json
import logging
import os
import sys
import re
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Dict, List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import torch
from crowdkit.aggregation import DawidSkene
from datasets import Dataset, DatasetDict
from matplotlib.figure import Figure

from .data_types import TaskType
from .dynamic_finetuner import DynamicTuner
from .utils import staticproperty

SOURCE_PATH = os.path.split(os.getcwd())[0]
sys.path.append(SOURCE_PATH)

DEVICE = (
    "cuda"
    if torch.cuda.is_available()
    else "mps"
    if torch.backends.mps.is_available()
    else "cpu"
)

os.environ["TOKENIZERS_PARALLELISM"] = "false"
os.environ["HYDRA_FULL_ERROR"] = "1"


log = logging.getLogger(__name__)

Report = Union[List[Dict[str, Union[int, List[float]]]], pd.DataFrame]


class AbstractSampleEvalMethod(ABC):
    def __init__(self, criteria: dict):
        self.criteria = criteria
        self.prepared_data: pd.DataFrame = None

    @abstractmethod
    def candidates_to_remove(
        self, dataset: DatasetDict, workdir: str, task_type: TaskType
    ) -> List[int]:
        raise NotImplementedError()

    @abstractmethod
    def save(self, savedir: str) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _preprocess_data(
        self, dataset: DatasetDict, workdir: str
    ) -> Dict[str, Dict[str, int | List[float]]]:
        raise NotImplementedError()

    @abstractmethod
    def _finalize_report(self, dataset: DatasetDict, report: Report) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _remove_prev_exps(self, workdir: str) -> None:
        raise NotImplementedError()

    @property
    def name(self):
        return re.sub("(?!^)([A-Z]+)", r"_\1", self.__class__.__name__).lower()

    @staticproperty
    @abstractmethod
    def need_tuned_model():
        raise NotImplementedError()

    @staticproperty
    @abstractmethod
    def need_tuner_instance():
        raise NotImplementedError()

    @staticproperty
    @abstractmethod
    def supported_tasks() -> List[TaskType]:
        raise NotImplementedError()


class BaseSampleEvalMethod(AbstractSampleEvalMethod):
    def save(self, savedir: str) -> None:
        savedir_postfix = os.path.join(savedir, self.name)
        os.makedirs(savedir_postfix, exist_ok=True)
        self._remove_prev_exps(savedir_postfix)

        path_file = os.path.join(savedir_postfix, f"{self.name}_by_sample_stat.csv")
        if "id" in self.prepared_data.columns:
            self.prepared_data = self.prepared_data.drop(columns=["id"])
        self.prepared_data.to_csv(path_file, index=False)
        log.info(f"{self.name} stats saved to {path_file}")

    def _remove_prev_exps(self, workdir: str) -> None:
        """removing previous data quality results"""
        filepaths = glob.glob(os.path.join(f"{workdir}", "**"), recursive=True)
        for filepath in filepaths:
            if os.path.isfile(filepath):
                os.remove(filepath)
                log.info(f"removed previous calculations: {filepath}")

    def _finalize_report(
        self,
        dataset: Dataset,
        report: Report,
        id2label: Dict[int, str],
    ) -> None:
        if isinstance(report, list):
            self.prepared_data = pd.DataFrame(report)
        elif isinstance(report, pd.DataFrame):
            self.prepared_data = report
        else:
            raise ValueError("report must be a list")

        if self.prepared_data.shape[0] == 0:
            return

        if "id_sample" not in self.prepared_data.columns:
            raise ValueError("id_sample column is missing in report")

        dataset_df = dataset.to_pandas()
        if "id" not in dataset_df.columns:
            dataset_df["id"] = dataset_df.index

        columns_from_dataset = ["id", "text"]
        if "text" not in dataset_df.columns:
            raise ValueError("text column is missing in dataset")

        if "pred_label" in self.prepared_data.columns:
            if self.prepared_data["pred_label"].dtype == int:
                self.prepared_data["pred_label"] = self.prepared_data["pred_label"].map(
                    id2label
                )
            elif self.prepared_data["pred_label"].dtype == list:
                self.prepared_data["pred_label"] = self.prepared_data[
                    "pred_label"
                ].apply(lambda x: list(map(id2label.get, x)))

        if "pred_label_by_epoch" in self.prepared_data.columns:
            self.prepared_data["pred_label_by_epoch"] = self.prepared_data[
                "pred_label_by_epoch"
            ].apply(lambda x: list(map(id2label.get, x)))

        if "tokens" in dataset_df.columns:
            columns_from_dataset.append("tokens")
        # token classification
        if "labels" in dataset_df.columns:
            columns_from_dataset.append("labels")
        # sequence classification
        elif "label" in dataset_df.columns:
            columns_from_dataset.append("label")

        dataset_df = dataset_df[columns_from_dataset]
        self.prepared_data = self.prepared_data.merge(
            dataset_df, left_on="id_sample", right_on="id", how="left"
        )
        if (
            "tokens" in dataset_df.columns
            and "id_token" in self.prepared_data
            and "labels" in dataset_df.columns
        ):

            def get_pred_tokens(row):
                row["tokens"] = row["tokens"][row["id_token"]]
                row["labels"] = row["labels"][row["id_token"]]
                return row

            self.prepared_data = self.prepared_data.apply(get_pred_tokens, axis=1)
            self.prepared_data.rename(
                columns={"tokens": "candidate_token"}, inplace=True
            )
            self.prepared_data.rename(columns={"labels": "label"}, inplace=True)
            self.prepared_data.drop(["id"], axis=1, inplace=True)

    def _task_resolver(self, task_type: TaskType):
        if task_type == TaskType.MULTICLASS:
            return self._candidates_multiclass
        if task_type == TaskType.NER:
            return self._candidates_ner

        raise ValueError(f"Unknown task type: {task_type}")

    def _candidates_multiclass(self, data: Dict[str, Dict[str, int | List[float]]]):
        raise NotImplementedError()

    def _candidates_ner(self, data: Dict[str, Dict[str, int | List[float]]]):
        raise NotImplementedError()


class ModelBasedMethod(BaseSampleEvalMethod):
    def candidates_to_remove(
        self,
        dataset: DatasetDict,
        workdir: str,
        task_type: TaskType,
        id2label: Optional[Dict[int, str]],
        tuner: DynamicTuner = None,
    ) -> List[int]:
        data = self._preprocess_data(dataset, workdir, tuner=tuner)

        task_method = self._task_resolver(task_type)
        prepared_data = task_method(data)

        self._finalize_report(dataset["train"], prepared_data, id2label)

        if self.prepared_data.shape[0] == 0:
            return []

        return self.prepared_data["id_sample"].tolist()

    @staticproperty
    def need_tuned_model():
        return True

    @staticproperty
    def need_tuner_instance():
        return False

    @staticproperty
    def supported_tasks() -> List[TaskType]:
        return [TaskType.MULTICLASS, TaskType.NER]


class LogitBasedMethod(ModelBasedMethod):
    need_last_epoch = True

    def __init__(
        self,
        criteria: dict,
    ):
        super().__init__(criteria)

    def _preprocess_data(
        self, dataset: DatasetDict, workdir: str, tuner: DynamicTuner = None
    ) -> Dict[str, Dict[str, int | List[float]]]:
        return self._load_data_from_tuner(workdir)

    def _load_data_from_tuner(
        self, path: str
    ) -> Dict[str, Dict[str, int | List[float]]]:
        """Load logits from saved result.jsonl. This file
        can be located in self.workdir as well as in subfolder

        Args:
            path (str): where result.jsonl located

        Returns:
            Dict[str, Dict[str, int | List[float]]]: reformated file for further processing
        """
        dirpattern = os.path.join(path, "result.jsonl")

        reformat_data = defaultdict(dict)

        with open(dirpattern, "r") as file:
            for line in file:
                record = json.loads(line)
                id_sample = record["id_sample"]
                epoch = record["epoch"]
                if id_sample not in reformat_data:
                    assert epoch == 0
                    reformat_data[id_sample]["gold"] = record["gold"]
                    reformat_data[id_sample]["logits"] = []
                if self.__class__.need_last_epoch:
                    reformat_data[id_sample]["logits"] = record["logits"]
                else:
                    reformat_data[id_sample]["logits"].append(record["logits"])

        return reformat_data


class Cartography(LogitBasedMethod):
    need_last_epoch = False

    def __init__(
        self,
        criteria: Dict[str, float] = {"confidence": 0.1, "variability": 0.1},
        include_ci: bool = False,
    ):
        """Sample Eval class for Dataset Cartography

        Args:
            workdir (List[str]|str): json files dirs
            criteria (dict): criterias to filter data
            include_ci (bool): Based on prior work on active bias (https://arxiv.org/abs/1704.07433).
                Default to False

        """
        super().__init__(criteria)
        self.include_ci = include_ci
        self.plot = None

    def _candidates_multiclass(self, data: dict) -> List[Dict[str, int | float]]:
        prepared_data = []

        log.info("Metrics computed: confidence, variability, correctness")

        for i, (id_sample, record) in enumerate(data.items()):
            row = {"id_sample": id_sample}

            true_probs_trend = []
            correctness_trend = []
            preds = []
            for epoch_logits in record["logits"]:
                probs = torch.nn.functional.softmax(torch.Tensor(epoch_logits), dim=-1)
                true_class_prob = float(probs[record["gold"]])
                true_probs_trend.append(true_class_prob)

                prediction = np.argmax(epoch_logits)
                is_correct = (prediction == record["gold"]).item()
                correctness_trend.append(is_correct)
                preds.append(prediction)

            row["correctness"] = sum(correctness_trend)
            row["confidence"] = np.mean(true_probs_trend)
            row["variability"] = self._variability(true_probs_trend)
            row["pred_label"] = preds[-1]
            row["pred_label_by_epoch"] = preds

            prepared_data.append(row)

        self.plot = self.__class__._plot_data_map(prepared_data)

        prepared_data = list(
            filter(
                lambda x: x["confidence"] <= self.criteria["confidence"]
                and x["variability"] <= self.criteria["variability"],
                prepared_data,
            )
        )

        return prepared_data

    def save(self, savedir: str) -> None:
        """savedir may differ from self.workdir
        self.workdir only for processing results/caching

        Args:
            savedir (str): savedir
        """
        savedir_postfix = os.path.join(savedir, self.name)
        os.makedirs(savedir_postfix, exist_ok=True)
        self._remove_prev_exps(savedir_postfix)
        path_file = os.path.join(savedir_postfix, "datamap.pdf")
        if self.plot is not None:
            self.plot.savefig(path_file, dpi=300)
        log.info(f"{self.name} plot saved to {path_file}")

        path_file = os.path.join(savedir_postfix, f"{self.name}_by_sample_stat.csv")
        if "id" in self.prepared_data.columns:
            self.prepared_data = self.prepared_data.drop(columns=["id"])
        self.prepared_data.to_csv(path_file, index=False)

    @staticmethod
    def _plot_data_map(
        prepared_data: List[dict],
        hue_metric: str = "correct.",
        show_hist: bool = True,
        max_instances_to_plot: int = 55000,
    ) -> Figure:
        prepared_data = pd.DataFrame(prepared_data)
        # Set style.
        sns.set_theme(style="whitegrid", font_scale=1.6, context="paper")
        log.info("Plotting...")

        # Subsample data to plot, so the plot is not too busy.
        dataframe = prepared_data.sample(
            n=(
                max_instances_to_plot
                if prepared_data.shape[0] > max_instances_to_plot
                else len(prepared_data)
            )
        )

        # Normalize correctness to a value between 0 and 1.
        dataframe = dataframe.assign(
            corr_frac=lambda d: d.correctness / d.correctness.max()
        )
        dataframe["correct."] = [f"{x:.1f}" for x in dataframe["corr_frac"]]

        main_metric = "variability"
        other_metric = "confidence"

        hue = hue_metric
        num_hues = len(dataframe[hue].unique().tolist())
        style = hue_metric if num_hues < 8 else None

        if not show_hist:
            fig, ax0 = plt.subplots(1, 1, figsize=(8, 6))
        else:
            fig = plt.figure(
                figsize=(14, 10),
            )
            gs = fig.add_gridspec(3, 2, width_ratios=[5, 1])
            ax0 = fig.add_subplot(gs[:, 0])

        # Make the scatterplot.
        # Choose a palette.
        pal = sns.diverging_palette(260, 15, n=num_hues, sep=10, center="dark")

        plot = sns.scatterplot(
            x=main_metric,
            y=other_metric,
            ax=ax0,
            data=dataframe,
            hue=hue,
            palette=pal,
            style=style,
            s=30,
        )

        # Annotate Regions.
        def bb(c):
            return dict(boxstyle="round,pad=0.3", ec=c, lw=2, fc="white")

        def func_annotate(text, xyc, bbc):
            return ax0.annotate(
                text,
                xy=xyc,
                xycoords="axes fraction",
                fontsize=15,
                color="black",
                va="center",
                ha="center",
                rotation=350,
                bbox=bb(bbc),
            )

        an1 = func_annotate("ambiguous", xyc=(0.9, 0.5), bbc="black")  # noqa
        an2 = func_annotate("easy-to-learn", xyc=(0.27, 0.85), bbc="r")  # noqa
        an3 = func_annotate("hard-to-learn", xyc=(0.35, 0.25), bbc="b")  # noqa

        if not show_hist:
            plot.legend(ncol=1, bbox_to_anchor=[0.175, 0.5], loc="right")
        else:
            plot.legend(fancybox=True, shadow=True, ncol=1)
        plot.set_xlabel("variability")
        plot.set_ylabel("confidence")

        if show_hist:
            plot.set_title("Data Map", fontsize=17)

            # Make the histograms.
            ax1 = fig.add_subplot(gs[0, 1])
            ax2 = fig.add_subplot(gs[1, 1])
            ax3 = fig.add_subplot(gs[2, 1])

            plott0 = dataframe.hist(column=["confidence"], ax=ax1, color="#622a87")
            plott0[0].set_title("")
            plott0[0].set_xlabel("confidence")
            plott0[0].set_ylabel("density")

            plott1 = dataframe.hist(column=["variability"], ax=ax2, color="teal")
            plott1[0].set_title("")
            plott1[0].set_xlabel("variability")
            plott1[0].set_ylabel("density")

            plot2 = sns.countplot(x="correct.", data=dataframe, ax=ax3, color="#86bf91")
            ax3.xaxis.grid(True)  # Show the vertical gridlines

            plot2.set_title("")
            plot2.set_xlabel("correctness")
            plot2.set_ylabel("density")

        fig.tight_layout()

        return fig

    def _variability(self, data: List[float | int]):
        # Functions to be applied to the data.
        if self.include_ci:
            variab = np.sqrt(
                np.var(data) + np.var(data) * np.var(data) / (len(data) - 1)
            )
            return variab

        return np.std(data)

    @staticproperty
    def supported_tasks() -> List[TaskType]:
        return [TaskType.MULTICLASS]


class VInfo(LogitBasedMethod):
    need_last_epoch = True

    def __init__(
        self,
        criteria: Dict[str, float] = {"v_info": 0.5},
    ):
        super().__init__(criteria)

    def _candidates_multiclass(
        self, data: dict
    ) -> dict:  # TODO: refactor format (need more general)
        proc_data = {}
        for stage in data:
            proc_data[stage] = self.__class__._v_info_prep(data[stage])

        prepared_data = proc_data["stage 0"]

        prepared_data["null_logits"] = proc_data["stage 1"]["logits"]
        prepared_data["score"] = prepared_data["null_logits"] - prepared_data["logits"]
        prepared_data = prepared_data.sort_values("score")
        prepared_data = prepared_data[prepared_data["score"] <= self.criteria["v_info"]]

        return prepared_data

    def _preprocess_data(
        self,
        dataset: DatasetDict,
        workdir: str,
        tuner: DynamicTuner,
    ) -> Dict[
        str, Dict[str, Dict[str, int | List[float]]]
    ]:  # TODO: need dataclasses/pydantic for complex types
        data_0 = self._load_data_from_tuner(workdir)

        # null model
        dataset = dataset.map(lambda _: {"text": " "})
        tuner.training_args.output_dir = os.path.join(workdir, "null_model")
        # train in tuner reload and reinit
        tuner.train(dataset["train"])

        data_1 = self._load_data_from_tuner(tuner.training_args.output_dir)

        preproc_data = {}
        preproc_data["stage 0"] = data_0
        preproc_data["stage 1"] = data_1

        return preproc_data

    @staticmethod
    def _v_info_prep(raw_data_stage: Dict[str, int | List[float]]) -> pd.DataFrame:
        def prob_gold(sample):
            sample["logits"] = sample["logits"][sample["gold"]]
            sample["logits"] = -torch.log2(sample["logits"]).item()

            return sample

        df = pd.DataFrame.from_dict(raw_data_stage, orient="index")
        df["logits"] = df["logits"].apply(
            lambda x: torch.softmax(torch.tensor(x), dim=-1)
        )
        df["pred_label"] = df["logits"].apply(lambda x: torch.argmax(x, dim=-1).item())
        df = df.apply(prob_gold, axis=1)
        df["id_sample"] = df.index
        df.drop("gold", axis=1, inplace=True)

        return df

    @staticproperty
    def need_tuner_instance():
        return True

    @staticproperty
    def supported_tasks() -> List[TaskType]:
        return [TaskType.MULTICLASS]


class Uncertainty(LogitBasedMethod):
    need_last_epoch = True

    def __init__(
        self,
        criteria: dict = {"uncertainty_threshold": 0.95},
    ):
        super().__init__(criteria)

    def _candidates_ner(
        self, data: Dict[int, List[float]]
    ) -> List[Dict[str, int | float]]:
        uncertainty_scores = []
        for idx in data:
            gold = data[idx]["gold"]
            logits = data[idx]["logits"]
            probs = torch.softmax(torch.tensor(logits), dim=-1)
            i = 0
            for _gold, _probs in zip(gold, probs):
                if _gold != -100:
                    row = {
                        "id_sample": idx,
                        "id_token": i,
                        "pred_label": _probs.argmax().item(),
                    }
                    uncertainty = 1 - _probs[_gold]
                    row["score"] = round(uncertainty.item(), 2)

                    uncertainty_scores.append(row)
                    i += 1

        uncertainty_scores = list(
            filter(
                lambda x: x["score"] > self.criteria["uncertainty_threshold"],
                uncertainty_scores,
            )
        )

        return uncertainty_scores

    def _candidates_multiclass(
        self, data: Dict[int, List[float]]
    ) -> List[Dict[str, int | float]]:
        uncertainty_scores = []
        for idx in data:
            gold = data[idx]["gold"]
            logits = data[idx]["logits"]
            probs = torch.softmax(torch.tensor(logits), dim=-1)
            row = {"id_sample": idx, "pred_label": probs.argmax().item()}
            uncertainty = 1 - probs[gold]
            row["score"] = round(uncertainty.item(), 2)
            uncertainty_scores.append(row)

        uncertainty_scores = list(
            filter(
                lambda x: x["score"] > self.criteria["uncertainty_threshold"],
                uncertainty_scores,
            )
        )

        return uncertainty_scores

    @staticproperty
    def supported_tasks() -> List[TaskType]:
        return [TaskType.NER, TaskType.MULTICLASS]


class Retag(LogitBasedMethod):
    need_last_epoch = True

    def __init__(
        self,
        criteria: dict = {},
    ):
        super().__init__(criteria)

    def _candidates_ner(
        self, data: Dict[int, List[float]]
    ) -> List[Dict[str, int | float]]:
        candidates = []
        for idx in data:
            gold = np.array(data[idx]["gold"])
            idx_true = np.where(gold != -100)[0]
            logits = np.array(data[idx]["logits"])

            # filter -100 tokens
            gold = gold[idx_true]
            logits = logits[idx_true]

            labels = logits.argmax(axis=-1).tolist()
            cols = np.where((gold != labels) & (gold != -100))[0].tolist()
            rows = [
                {"id_sample": idx, "id_token": col, "pred_label": label}
                for col, label in zip(cols, labels)
            ]
            candidates.extend(rows)

        return candidates

    def _candidates_multiclass(self, data) -> List[Dict[str, int | float]]:
        candidates = []
        for idx in data:
            gold = data[idx]["gold"]
            logits = data[idx]["logits"]
            label = np.array(logits).argmax(axis=-1)
            if gold != label:
                row = {"id_sample": idx, "pred_label": label}
                candidates.append(row)

        return candidates

    @staticproperty
    def supported_tasks() -> List[TaskType]:
        return [TaskType.NER, TaskType.MULTICLASS]


class LabelAggregation(ModelBasedMethod):
    def __init__(
        self,
        criteria: dict = {},
        num_repetitions: int = 10,
    ):
        super().__init__(criteria)
        self.num_repetitions = num_repetitions

    # TODO: Label Aggregation for sequence classification
    def candidates_to_remove(
        self,
        dataset: DatasetDict,
        workdir: str,
        task_type: TaskType,
        id2label: Optional[Dict[int, str]] = None,
        tuner: DynamicTuner = None,
    ) -> List[int]:
        raw_data, ground_truth = self._preprocess_data(dataset, workdir, tuner=tuner)
        ground_truth_flat = ground_truth.reshape(-1)
        # (num_rep, N, max_length, num_classes)
        mc_dropout_resulit = np.array(raw_data)
        # (num_rep, N, max_length)
        label_result = mc_dropout_resulit.argmax(axis=-1)
        # (num_rep, N*max_length)
        label_result = label_result.reshape(label_result.shape[0], -1)
        # (N*max_length, num_rep)
        label_result = label_result.transpose()

        tasks = np.repeat(np.arange(label_result.shape[0]), self.num_repetitions)
        performers = np.tile(np.arange(self.num_repetitions), label_result.shape[0])
        labels = label_result.reshape(-1)

        assert len(labels) == tasks.shape[0] == performers.shape[0]

        data = {"task": tasks, "worker": performers, "label": labels}
        df = pd.DataFrame(data)
        log.info("Started aggregation model")
        log.info(f"size for DawidSkene model: {df.shape}")
        dawid_skene_model = DawidSkene(n_iter=100)
        encoded_predictions = dawid_skene_model.fit_predict(df)
        encoded_predictions = encoded_predictions.to_numpy()
        log.info("Finished aggregation model")

        flags_condition = (encoded_predictions != ground_truth_flat) & (
            ground_truth_flat != -100
        )
        # flags_condition = (encoded_predictions[ground_truth!=-100] != ground_truth[ground_truth!=-100])
        flags_flat = np.where(flags_condition)[0]

        if flags_flat.shape[0] == 0:
            self._finalize_report(dataset["train"], [], id2label)
            return []

        N, token_n = np.unravel_index(flags_flat, ground_truth.shape)
        flags = list(zip(N, token_n))

        prepared_data = []
        # disalignment
        for i, flag in enumerate(flags):
            m, n = flag
            correction = sum(ground_truth[m][:n] == -100)
            row = {
                "id_sample": m,
                "id_token": n - correction,
                "pred_label": encoded_predictions[(m * ground_truth.shape[1]) + n],
            }
            prepared_data.append(row)

        self._finalize_report(dataset["train"], prepared_data, id2label)

        return self.prepared_data["id_sample"].tolist()

    def _preprocess_data(
        self, dataset: DatasetDict, workdir: str, tuner: DynamicTuner
    ) -> Dict[str, Dict[str, int | List[float]]]:
        raw_data = tuner.mc_droupout(
            dataset["train"], num_repetitions=self.num_repetitions
        )

        return raw_data

    @staticproperty
    def need_tuned_model():
        return True

    @staticproperty
    def need_tuner_instance():
        return True

    @staticproperty
    def supported_tasks() -> List[TaskType]:
        return [TaskType.NER]



MAX_ANC_SET_FIT_SIZE = 5
MAX_SET_FIT_SIZE = 80
MIN_ANC_SET_FIT_SIZE = 2
LOWER_OPTIMAL_SETFIT_SIZE = 70
LOW_RES_CLS_FRACT_THRESHOLD = 0.3
SETFIT_NAME = "setfit"
ANCSETFIT_NAME = "ancsetfit"
FINETUNING_NAME = "finetuning"
MODEL_SAVE_PATH = "models"
TOKEN_CLF_FINETUNING = "token_clf_finetuning"
OOD_METHOD_MAP = {
    SETFIT_NAME: "SetFitOOD",
    ANCSETFIT_NAME: "AncSetFitMethod",
    FINETUNING_NAME: "FinetunerWithOOD",
    TOKEN_CLF_FINETUNING: "TokenClassificationFinetuner",
}
NO_OOD_METHOD_MAP = {
    SETFIT_NAME: "SetFitMethod",
    ANCSETFIT_NAME: "AncSetFitMethod",
    FINETUNING_NAME: "Finetuner",
    TOKEN_CLF_FINETUNING: "TokenClassificationFinetuner",
}

ALL_AVAILABLE_METHODS = [
    "AncSetFitMethod",
    "Finetuner",
    "SetFitMethod",
    "TokenClassificationFinetuner",
    "FinetunerWithOOD",
    "SetFitOOD",
    "AncSetFitOOD",
]

METHOD_HUMAN_READABLE_NAME_MAP = {
    "AncSetFitMethod": "AncSetFit",
    "AncSetFitOOD": "AncSetFit",
    "SetFitMethod": "SetFit",
    "SetFitOOD": "SetFit",
    "Finetuner": "Finetuning",
    "TokenClassificationFinetuner": "Finetuning",
    "FinetunerWithOOD": "Finetuning",
}

TRAIN_KEY = "train"
DEV_KEY = "dev"
TEST_KEY = "test"
UPSAMPLE_AUGMENTEX = "augmentex"
UPSAMPLE_LLM = "llm"
LLM_AUGMENTATION_THRESHOLD = 81

DATA_LOADING_STAGE = "loading data"
DATA_PROCESSING_STAGE = "processing data"
DATA_EVALUATION_STAGE = "evaluating data"
HPO_STAGE = "searching optimal hyperparameters"
PRETRAINING_STAGE = "pretraining"
TRAINING_STAGE = "training"
DEV_EVALUATION_STAGE = "evaluating the model"
SAVING_STAGE = "saving the model"
SPAN_DETECTION_ADAPTATION_STAGE = "span detection adaptation"
DQ_TRAIN = "scanning data"



from typing import Union

import numpy as np
import torch
from torch import nn
from sklearn.metrics import roc_curve


def is_normalized(logits, tol=1e-5):
    """
    Check if the logits vector is normalized (i.e., softmax has been applied).

    Args:
        logits (torch.Tensor): The logits vector to check.
        tol (float): Tolerance for floating-point comparison.

    Returns:
        bool: True if the logits are normalized, False otherwise.
    """
    # Check if all values are between 0 and 1
    in_range = torch.all((logits >= 0) & (logits <= 1)).item()

    # Check if the sum of the logits is approximately 1
    sum_close_to_one = torch.isclose(
        logits.sum(), torch.tensor(1.0, dtype=logits.dtype), atol=tol
    ).item()

    return in_range and sum_close_to_one


class MaxSoftmaxProbabilityDetector(nn.Module):
    """Maximum Softmax Probability (MSP) OOD detector.

    Detects OOD samples based on the maximum softmax probability.
    Samples with max probability below the threshold are classified as OOD.
    """

    def __init__(
        self,
        threshold: Union[np.float32, float],
    ):
        """Initialize the MSP detector.

        Args:
            threshold: Probability threshold for in-scope classification.
        """
        super().__init__()
        self.threshold = nn.Parameter(torch.Tensor([threshold]), requires_grad=False)

    def forward(self, scores: torch.Tensor) -> torch.Tensor:
        """Detect OOD samples.

        Args:
            scores: Batch of logits or probabilities.

        Returns:
            Tensor of 0/1 values where 1 indicates OOD.
        """
        if not is_normalized(scores[0]):
            scores = torch.softmax(scores, dim=1)
        max_probs = scores.max(dim=1).values
        return torch.where(max_probs > self.threshold, 0.0, 1.0)  # OOD-1, ID-0


class MaxSoftmaxProbabilityTrainer:
    """Trainer for Maximum Softmax Probability OOD detector.

    Determines optimal threshold using Youden's J statistic on
    in-scope and OOS validation samples.
    """

    def __init__(self, *args, **kwargs):
        pass

    @staticmethod
    def train(
        inscope_logits: Union[np.ndarray, torch.tensor],
        oos_logits: Union[np.ndarray, torch.tensor],
        *args,
        **kwargs,
    ) -> MaxSoftmaxProbabilityDetector:
        """Train the MSP detector.

        Args:
            inscope_logits: Logits from in-scope validation samples.
            oos_logits: Logits from OOS validation samples.

        Returns:
            Trained MaxSoftmaxProbabilityDetector with optimal threshold.
        """
        inscope_logits = torch.tensor(inscope_logits)
        oos_logits = torch.tensor(oos_logits)
        if not is_normalized(inscope_logits[0]):
            inscope_logits = torch.softmax(torch.tensor(inscope_logits), dim=1)
        inscope_scores = np.array(inscope_logits).max(axis=1)
        if not is_normalized(oos_logits[0]):
            oos_logits = torch.softmax(torch.tensor(oos_logits), dim=1)
        oos_scores = np.array(oos_logits).max(axis=1)
        scores = np.hstack([inscope_scores, oos_scores])
        labels = np.hstack([np.ones_like(inscope_scores), np.zeros_like(oos_scores)])
        fpr, tpr, thresholds = roc_curve(labels, scores)
        # Youden's J Statistic balances sensitivity (TPR) and specificity (1 - FPR)
        # The threshold that maximizes Youden's J is considered optimal
        youden_j = tpr - fpr
        optimal_idx = np.argmax(youden_j)
        optimal_threshold = thresholds[optimal_idx]
        return MaxSoftmaxProbabilityDetector(threshold=optimal_threshold)  # type: ignore



from typing import Any, Dict, Optional
import torch
from transformers import AutoTokenizer
from collections import defaultdict
import datasets


def batch(iterable, size=1):
    iterable_length = len(iterable)
    for idx in range(0, iterable_length, size):
        yield iterable[idx : min(idx + size, iterable_length)]


def initialize_transformers_tokenizer(
    model_name_or_path: str,
    max_seq_length: int,
    tokenizer_kwargs: Optional[Dict[str, Any]] = None,
):
    default_tokenizer_kwargs = {"model_max_length": max_seq_length}
    if not tokenizer_kwargs:
        tokenizer_kwargs = default_tokenizer_kwargs
    else:
        default_tokenizer_kwargs.update(tokenizer_kwargs)
        tokenizer_kwargs = default_tokenizer_kwargs
    return AutoTokenizer.from_pretrained(
        model_name_or_path, use_fast=True, **tokenizer_kwargs
    )


def resolve_device():
    if torch.cuda.is_available():
        device = "cuda"
    elif torch.backends.mps.is_available():
        device = "mps"
    else:
        device = "cpu"
    return device


def sample_ner_dataset(dataset: datasets.Dataset, n_shot: int) -> datasets.Dataset:
    """
    This function should take in a datasets.Dataset with columns
    text (a string with text) and labels (a list of BIO labels)
    and return a datasets.Dataset with same structure but each
    entity type should occur approximately n_shot times. All
    empty (all BIO labels are O) samples should be excluded.
    """

    # Filter out examples with all 'O' labels
    def has_entities(example):
        return any(label != "O" for label in example["labels"])

    filtered_dataset = dataset.filter(has_entities)

    # If no examples left after filtering, return empty dataset
    if len(filtered_dataset) == 0:
        return datasets.Dataset.from_dict({"text": [], "labels": []})

    # Compute entity counts for each example
    def compute_entity_counts(example):
        labels = example["labels"]
        entity_counts = defaultdict(int)
        current_entity = None
        for label in labels:
            if label.startswith("B-"):
                entity_type = label[2:]
                entity_counts[entity_type] += 1
                current_entity = entity_type
            elif label.startswith("I-"):
                entity_type = label[2:]
                if current_entity == entity_type:
                    continue  # Part of the same entity
                else:
                    current_entity = None  # Invalid I- tag, treat as O
            else:
                current_entity = None
        # Convert to regular dict for serialization
        example["entity_counts"] = dict(entity_counts)
        return example

    def normalize_dicts(examples):
        new_data = {"entity_counts": []}
        for counts in examples["entity_counts"]:
            new_data["entity_counts"].append(
                {
                    key: value if value is not None else 0
                    for key, value in counts.items()
                }
            )
        return new_data

    filtered_dataset = filtered_dataset.map(compute_entity_counts).map(
        normalize_dicts, batched=True
    )
    # return filtered_dataset

    # Collect all entity types present in the dataset
    entity_types = list(filtered_dataset[0]["entity_counts"].keys())
    # for example in filtered_dataset:
    #     entity_types.update(example['entity_counts'].keys())
    # entity_types = list(entity_types)

    if not entity_types:
        return datasets.Dataset.from_dict({"text": [], "labels": []})

    # Shuffle entity types to randomize processing order
    # random.shuffle(entity_types)
    print(entity_types)

    # Initialize remaining counts for each entity type
    remaining_counts = {et: n_shot for et in entity_types}
    selected_indices = set()

    # Process each entity type to collect examples
    for entity_type in entity_types:
        # Collect candidate examples that have the current entity type and are not selected yet
        candidates = []
        for idx in range(len(filtered_dataset)):
            if idx not in selected_indices:
                example = filtered_dataset[idx]
                if entity_type in example["entity_counts"]:
                    count = example["entity_counts"][entity_type]
                    candidates.append((idx, count))

        # Sort candidates by count descending to maximize contribution per example
        try:
            candidates.sort(key=lambda x: x[1], reverse=True)
        except Exception as e:
            print(candidates)
            raise (e)

        for idx, count in candidates:
            if remaining_counts[entity_type] <= 0:
                break
            if idx not in selected_indices:
                selected_indices.add(idx)
                # Update remaining counts for all entity types in this example
                example = filtered_dataset[idx]
                for et, cnt in example["entity_counts"].items():
                    if et in remaining_counts:
                        remaining_counts[et] = max(0, remaining_counts[et] - cnt)

    # Create the sampled dataset
    selected_indices = sorted(selected_indices)
    sampled_dataset = filtered_dataset.select(selected_indices)

    # Remove the temporary 'entity_counts' column
    sampled_dataset = sampled_dataset.remove_columns("entity_counts")

    return sampled_dataset



import gc
import json
import os
import shutil
import threading
import time
import uuid
from collections import deque
from dataclasses import dataclass
from datetime import timedelta
from pathlib import Path
from typing import List, Optional, Dict, Any
import numpy as np
import pandas as pd
import streamlit as st
from datasets import DatasetDict
from dotenv import load_dotenv
import torch

load_dotenv()

from open_autonlu.auto_classes import (  # noqa: E402
    TextClassificationInferenceManager,
    TextClassificationTrainingPipeline,
    TokenClassificationTrainingPipeline,
    TokenClassificationInferenceManager,
)

from open_autonlu.data.utils import convert_bio_to_spans  # noqa: E402
from open_autonlu.methods.data_types import (  # noqa: E402
    ClassificationEvaluationResult,
    SaveFormat,
    OodMethod,
    TrainingArtifactInfo,
)

import plotly.express as px  # noqa: E402

DATA_DIR = "data"
MODEL_BASE_DIR = "saved_models/streamlit_demo"
MAX_INFERENCE_BATCH_SIZE = 20000

_DEFAULT_TRAINING_CONFIG: Dict[str, Any] = {
    "ood_method": "AUTO",
    "threshold_factor": 1.0,
    "llm_augmentation": False,
    "llm_test_generation": False,
}


def _default_llm_config() -> Dict[str, str]:
    return {
        "api_key": os.environ.get("MODEL_API_KEY", ""),
        "base_url": os.environ.get("BASE_URL", ""),
        "model_id": os.environ.get("MODEL_ID", ""),
    }


# Process scheduler state shared across all reruns
@st.cache_resource
def _shared_scheduler_state():
    return {
        "heavy_op_lock": threading.Lock(),
        "queue_lock": threading.Lock(),
        "queue": deque(),
    }


_SCHED = _shared_scheduler_state()
_heavy_op_lock: threading.Lock = _SCHED["heavy_op_lock"]
_queue_lock: threading.Lock = _SCHED["queue_lock"]
_heavy_op_queue: deque = _SCHED["queue"]

_QUEUE_REFRESH_SECONDS = 1


_SESSION_META_FILENAME = "session_meta.json"


def _session_id() -> str:
    if "session_id" in st.session_state:
        return st.session_state.session_id
    sid = st.query_params.get("sid")
    if sid and len(sid) == 12 and all(c in "0123456789abcdef" for c in sid.lower()):
        st.session_state.session_id = sid
        return sid
    new_id = uuid.uuid4().hex[:12]
    st.session_state.session_id = new_id
    st.query_params["sid"] = new_id
    return new_id


def _current_queue_position() -> Optional[int]:
    sid = _session_id()
    with _queue_lock:
        for i, (s, _) in enumerate(_heavy_op_queue, 1):
            if s == sid:
                return i
    return None


def _session_data_dir() -> str:
    return os.path.join(DATA_DIR, _session_id())


def _session_model_dir() -> str:
    return os.path.join(MODEL_BASE_DIR, _session_id())


def get_train_path() -> str:
    return os.path.join(_session_data_dir(), "demo_train.csv")


def get_test_path() -> str:
    return os.path.join(_session_data_dir(), "demo_test.csv")


def get_ner_train_path() -> str:
    return os.path.join(_session_data_dir(), "demo_train.json")


def get_ner_test_path() -> str:
    return os.path.join(_session_data_dir(), "demo_test.json")


_PROJECT_ROOT = Path(__file__).resolve().parent
_SAMPLE_DATA_ROOT = _PROJECT_ROOT / "examples" / "test_data"
SAMPLE_TRAIN_TEXT = (
    _SAMPLE_DATA_ROOT / "noise_n_shot_data" / "snips_200_300_0.01_42_train.csv"
)
SAMPLE_TEST_TEXT = (
    _SAMPLE_DATA_ROOT / "noise_n_shot_data" / "snips_200_300_0.01_42_test.csv"
)
SAMPLE_TRAIN_NER = _SAMPLE_DATA_ROOT / "noise_n_shot_data_ner" / "train.json"
SAMPLE_TEST_NER = _SAMPLE_DATA_ROOT / "noise_n_shot_data_ner" / "test.json"

AGGREGATE_ROWS = frozenset(["accuracy", "macro avg", "weighted avg", "micro avg"])

TASK_TEXT_CLF = "Text Classification"
TASK_NER = "Named Entity Recognition"


def ensure_dirs():
    os.makedirs(_session_data_dir(), exist_ok=True)
    os.makedirs(_session_model_dir(), exist_ok=True)


def _session_meta_path() -> str:
    return os.path.join(_session_data_dir(), _SESSION_META_FILENAME)


def _update_meta_model_path(meta_path: str, model_path: str) -> None:
    try:
        with open(meta_path, "r", encoding="utf-8") as f:
            meta = json.load(f)
    except Exception:
        meta = {}
    meta["model_path"] = model_path
    with open(meta_path, "w", encoding="utf-8") as f:
        json.dump(meta, f, indent=2)


_JOB_RUNNING_FILENAME = "job_running.json"


def _job_running_path() -> str:
    return os.path.join(_session_data_dir(), _JOB_RUNNING_FILENAME)


def _save_session_meta():
    ensure_dirs()
    meta = {
        "task_type": st.session_state.get("task_type", TASK_TEXT_CLF),
        "language": st.session_state.get("language", "en"),
        "tab_choice": st.session_state.get("tab_choice", "Data Input"),
        "model_path": st.session_state.get("model_path"),
        "training_config": st.session_state.get("training_config", {}),
        "llm_config": st.session_state.get("llm_config", {}),
    }
    with open(_session_meta_path(), "w", encoding="utf-8") as f:
        json.dump(meta, f, indent=2)


def _load_session_meta() -> Optional[Dict[str, Any]]:
    path = _session_meta_path()
    if not os.path.isfile(path):
        return None
    try:
        with open(path, encoding="utf-8") as f:
            return json.load(f)
    except Exception:
        return None


def _find_latest_model_dir() -> Optional[str]:
    d = _session_model_dir()
    if not os.path.isdir(d):
        return None
    best_path = None
    best_mtime = 0
    for name in os.listdir(d):
        path = os.path.join(d, name)
        if os.path.isdir(path):
            m = os.path.getmtime(path)
            if m > best_mtime:
                best_mtime = m
                best_path = path
    return best_path


_TRAINING_RESULTS_FILENAME = "training_results.json"
_DQ_RESULTS_FILENAME = "dq_results.json"
_DQ_REPORT_FILENAME = "dq_aggregated_report.csv"


@dataclass
class _PersistedDqOutput:
    indices_to_remove: List[int]
    aggregated_report_path: Optional[str]
    splits_filtered: Any = None


def _save_training_results(
    artifact, method_name: str, path: Optional[str] = None
) -> None:
    if artifact is None:
        return
    if path is None:
        ensure_dirs()
        path = os.path.join(_session_data_dir(), _TRAINING_RESULTS_FILENAME)
    try:
        labels = getattr(artifact, "labels", []) or []
        payload = {
            "method_name": method_name or "unknown",
            "labels": labels,
        }
        if artifact.test_metrics is not None:
            cm = artifact.test_metrics.confusion_matrix
            cm_list = cm.tolist() if hasattr(cm, "tolist") else None
            payload["classification_report"] = getattr(
                artifact.test_metrics, "classification_report", {}
            )
            payload["f1"] = getattr(artifact.test_metrics, "f1", None)
            payload["confusion_matrix"] = cm_list
        with open(path, "w", encoding="utf-8") as f:
            json.dump(payload, f, indent=2)
    except Exception:
        pass


def _load_training_results() -> Optional[tuple]:
    path = os.path.join(_session_data_dir(), _TRAINING_RESULTS_FILENAME)
    if not os.path.isfile(path):
        return None
    try:
        with open(path, encoding="utf-8") as f:
            data = json.load(f)
        method_name = data.get("method_name", "unknown")
        labels = data.get("labels", [])
        classification_report = data.get("classification_report", {})
        f1 = data.get("f1")
        cm_list = data.get("confusion_matrix")
        test_metrics = None
        if classification_report or f1 is not None or cm_list:
            cm = np.array(cm_list) if cm_list else np.array([])
            test_metrics = ClassificationEvaluationResult(
                confusion_matrix=cm,
                classification_report=classification_report or {},
                f1=float(f1) if f1 is not None else 0.0,
            )
        artifact = TrainingArtifactInfo(
            hyperparameters={},
            labels=labels,
            dev_metrics=None,
            test_metrics=test_metrics,
            test_metrics_inscope=None,
        )
        return (artifact, method_name)
    except Exception:
        return None


def _save_dq_results(dq_output, data_dir: Optional[str] = None) -> None:
    if dq_output is None:
        return
    if data_dir is None:
        ensure_dirs()
        data_dir = _session_data_dir()
    indices = list(getattr(dq_output, "indices_to_remove", None) or [])
    src_report = getattr(dq_output, "aggregated_report_path", None)
    dst_report = os.path.join(data_dir, _DQ_REPORT_FILENAME)
    if src_report and os.path.isfile(src_report):
        try:
            shutil.copy2(src_report, dst_report)
        except Exception:
            dst_report = None
    else:
        dst_report = None
    path = os.path.join(data_dir, _DQ_RESULTS_FILENAME)
    try:
        with open(path, "w", encoding="utf-8") as f:
            json.dump(
                {"indices_to_remove": indices, "aggregated_report_path": dst_report},
                f,
                indent=2,
            )
    except Exception:
        pass


def _load_dq_results() -> Optional[_PersistedDqOutput]:
    path = os.path.join(_session_data_dir(), _DQ_RESULTS_FILENAME)
    if not os.path.isfile(path):
        return None
    try:
        with open(path, encoding="utf-8") as f:
            data = json.load(f)
        indices = data.get("indices_to_remove", [])
        report_path = data.get("aggregated_report_path")
        if report_path and not os.path.isfile(report_path):
            report_path = None
        return _PersistedDqOutput(
            indices_to_remove=indices,
            aggregated_report_path=report_path,
            splits_filtered=None,
        )
    except Exception:
        return None


def empty_cache():
    if torch.backends.mps.is_available():
        torch.mps.empty_cache()
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
    gc.collect()


def init_state():
    if "train_data" in st.session_state:
        _session_id()
        if st.session_state.get("training_artifact") is None:
            loaded = _load_training_results()
            if loaded is not None:
                st.session_state.training_artifact, st.session_state.method_name = (
                    loaded[0],
                    loaded[1],
                )
        if st.session_state.get("dq_output") is None:
            loaded_dq = _load_dq_results()
            if loaded_dq is not None:
                st.session_state.dq_output = loaded_dq
        if st.session_state.get("model_path") is None:
            meta = _load_session_meta()
            model_path = meta.get("model_path") if meta else None
            if model_path and os.path.isdir(model_path):
                st.session_state.model_path = model_path
            else:
                st.session_state.model_path = _find_latest_model_dir()
        return
    _session_id()
    meta = _load_session_meta()
    if meta is not None:
        # Restore data from disk
        st.session_state.task_type = meta.get("task_type", TASK_TEXT_CLF)
        st.session_state.language = meta.get("language", "en")
        st.session_state.tab_choice = meta.get("tab_choice", "Data Input")
        st.session_state.training_config = meta.get(
            "training_config", dict(_DEFAULT_TRAINING_CONFIG)
        )
        st.session_state.llm_config = meta.get("llm_config", _default_llm_config())
        model_path = meta.get("model_path")
        if model_path and os.path.isdir(model_path):
            st.session_state.model_path = model_path
        else:
            st.session_state.model_path = _find_latest_model_dir()
        job_running_path = _job_running_path()
        if os.path.isfile(job_running_path):
            try:
                with open(job_running_path, encoding="utf-8") as f:
                    job_info = json.load(f)
                action = job_info.get("action", "training")
                st.session_state.tab_choice = (
                    "Training" if action == "training" else "Data Quality"
                )
                st.session_state.training_in_progress = True
            except Exception:
                st.session_state.training_in_progress = False
        else:
            st.session_state.training_in_progress = False
        st.session_state.train_data = pd.DataFrame(columns=["text", "label"])
        st.session_state.test_data = None
        st.session_state.ner_train_data = None
        st.session_state.ner_test_data = None
        if os.path.isfile(get_train_path()):
            try:
                st.session_state.train_data = pd.read_csv(get_train_path())
                st.session_state.train_data = normalize_train_data(
                    st.session_state.train_data
                )
            except Exception:
                pass
        if os.path.isfile(get_test_path()):
            try:
                st.session_state.test_data = pd.read_csv(get_test_path())
                st.session_state.test_data = normalize_test_data(
                    st.session_state.test_data
                )
            except Exception:
                pass
        if os.path.isfile(get_ner_train_path()):
            try:
                with open(get_ner_train_path(), encoding="utf-8") as f:
                    st.session_state.ner_train_data = json.load(f)
            except Exception:
                pass
        if os.path.isfile(get_ner_test_path()):
            try:
                with open(get_ner_test_path(), encoding="utf-8") as f:
                    st.session_state.ner_test_data = json.load(f)
            except Exception:
                pass
        loaded_dq = _load_dq_results()
        st.session_state.dq_output = loaded_dq if loaded_dq is not None else None
        loaded_training = _load_training_results()
        if loaded_training is not None:
            st.session_state.training_artifact, st.session_state.method_name = (
                loaded_training[0],
                loaded_training[1],
            )
        else:
            st.session_state.training_artifact = None
            st.session_state.method_name = None
        st.session_state.batch_predictions = None
        return
    # Default empty state
    st.session_state.task_type = TASK_TEXT_CLF
    st.session_state.train_data = pd.DataFrame(columns=["text", "label"])
    st.session_state.test_data = None
    st.session_state.ner_train_data = None
    st.session_state.ner_test_data = None
    st.session_state.dq_output = None
    st.session_state.training_artifact = None
    st.session_state.method_name = None
    st.session_state.model_path = None
    st.session_state.batch_predictions = None
    st.session_state.training_config = dict(_DEFAULT_TRAINING_CONFIG)
    st.session_state.llm_config = _default_llm_config()
    st.session_state.language = "en"
    st.session_state.tab_choice = "Data Input"
    st.session_state.training_in_progress = False


def on_task_type_change():
    st.session_state.dq_output = None
    st.session_state.training_artifact = None
    st.session_state.method_name = None
    st.session_state.model_path = None
    st.session_state.batch_predictions = None
    _save_session_meta()


def is_ner() -> bool:
    return st.session_state.task_type == TASK_NER


# Data normalization


def normalize_train_data(df: pd.DataFrame) -> Optional[pd.DataFrame]:
    if "text" not in df.columns or "label" not in df.columns:
        st.error("Train data should contain text and label fields")
        return None
    if "anc_label" in df.columns:
        anc_series = df["anc_label"]
        if anc_series.isna().all() or (anc_series.astype(str).str.strip() == "").all():
            df = df.drop(columns=["anc_label"])
    columns = ["text", "label"]
    if "anc_label" in df.columns:
        columns.append("anc_label")
    return df[columns]


def normalize_test_data(df: pd.DataFrame) -> Optional[pd.DataFrame]:
    if "text" not in df.columns or "label" not in df.columns:
        st.error("Test data should contain text and label fields")
        return None
    return df[["text", "label"]]


# Data input callbacks


def add_train_file():
    if st.session_state.train_file is None:
        return
    df = pd.read_csv(st.session_state.train_file)
    df = normalize_train_data(df)
    if df is None:
        return
    st.session_state.train_data = df
    save_inputs()


def add_test_file():
    if st.session_state.test_file is None:
        return
    df = pd.read_csv(st.session_state.test_file)
    df = normalize_test_data(df)
    if df is None:
        return
    st.session_state.test_data = df
    save_inputs()


def add_ner_train_file():
    if st.session_state.ner_train_file is None:
        return
    try:
        content = json.load(st.session_state.ner_train_file)
        if not isinstance(content, list):
            st.error("JSON should contain a list of records")
            return
        st.session_state.ner_train_data = content
        save_ner_inputs()
    except json.JSONDecodeError:
        st.error("Invalid JSON file")


def add_ner_test_file():
    if st.session_state.ner_test_file is None:
        return
    try:
        content = json.load(st.session_state.ner_test_file)
        if not isinstance(content, list):
            st.error("JSON should contain a list of records")
            return
        st.session_state.ner_test_data = content
        save_ner_inputs()
    except json.JSONDecodeError:
        st.error("Invalid JSON file")


# Save helpers


def save_inputs():
    ensure_dirs()
    st.session_state.train_data.to_csv(get_train_path(), index=False)
    if st.session_state.test_data is not None:
        st.session_state.test_data.to_csv(get_test_path(), index=False)
    elif os.path.exists(get_test_path()):
        os.remove(get_test_path())
    _save_session_meta()


def save_ner_inputs():
    ensure_dirs()
    if st.session_state.ner_train_data is not None:
        with open(get_ner_train_path(), "w", encoding="utf-8") as f:
            json.dump(st.session_state.ner_train_data, f, ensure_ascii=False)
    if st.session_state.ner_test_data is not None:
        with open(get_ner_test_path(), "w", encoding="utf-8") as f:
            json.dump(st.session_state.ner_test_data, f, ensure_ascii=False)
    elif os.path.exists(get_ner_test_path()):
        os.remove(get_ner_test_path())
    _save_session_meta()


# NER data preview


def ner_data_summary(data: list) -> pd.DataFrame:
    rows = []
    for i, record in enumerate(data):
        text = record.get("text", "")
        spans = record.get("spans", [])
        entity_labels = sorted({s.get("label", "?") for s in spans}) if spans else []
        rows.append(
            {
                "#": i,
                "text": text[:80] + "..." if len(text) > 80 else text,
                "entities": len(spans),
                "labels": ", ".join(entity_labels) if entity_labels else "—",
            }
        )
    return pd.DataFrame(rows)


# Data Quality


def _run_dq_impl(job_path: str):
    if is_ner():
        _run_ner_dq(job_path)
    else:
        _run_text_clf_dq(job_path)


def run_dq():
    sid = _session_id()
    with _queue_lock:
        if any(s == sid for s, _ in _heavy_op_queue):
            return
        _heavy_op_queue.append((sid, "dq"))
    st.session_state.queued_action = "dq"
    st.session_state.tab_choice = "Data Quality"
    _save_session_meta()


def _run_text_clf_dq(job_path: str):
    if st.session_state.train_data.empty:
        st.error("Training data is missing")
        return
    save_inputs()
    _save_session_meta()
    lang = st.session_state.get("language", "en")
    train_path = get_train_path()
    test_path = get_test_path() if st.session_state.test_data is not None else None
    data_dir = _session_data_dir()
    ensure_dirs()

    worker_result = [None]
    worker_error = [None]

    def _worker():
        try:
            pipeline = TextClassificationTrainingPipeline(
                train_path=train_path,
                test_path=test_path,
                config_overrides={"language": lang},
            )
            dq_output = pipeline.diagnose()
            if dq_output is not None:
                _save_dq_results(dq_output, data_dir=data_dir)
            worker_result[0] = dq_output
        except Exception as e:
            worker_error[0] = e
        finally:
            _heavy_op_lock.release()
            try:
                if os.path.isfile(job_path):
                    os.remove(job_path)
            except Exception:
                pass

    t = threading.Thread(target=_worker)
    t.start()
    try:
        t.join()
    except BaseException:
        return

    if worker_error[0] is not None:
        st.error(f"Error during data quality check: {worker_error[0]}")
        st.session_state.dq_output = None
        empty_cache()
        return

    st.session_state.dq_output = worker_result[0]
    empty_cache()
    if worker_result[0] is None:
        st.warning(
            "Data quality check skipped: dataset is too small for quality evaluation."
        )


def _run_ner_dq(job_path: str):
    if st.session_state.ner_train_data is None:
        st.error("NER training data is missing")
        return
    save_ner_inputs()
    lang = st.session_state.get("language", "en")
    ner_train_path = get_ner_train_path()
    ner_test_path = (
        get_ner_test_path() if st.session_state.ner_test_data is not None else None
    )
    data_dir = _session_data_dir()
    ensure_dirs()

    worker_result = [None]
    worker_error = [None]

    def _worker():
        try:
            pipeline = TokenClassificationTrainingPipeline(
                train_path=ner_train_path,
                test_path=ner_test_path,
                config_overrides={"language": lang},
            )
            dq_output = pipeline.diagnose()
            if dq_output is not None:
                _save_dq_results(dq_output, data_dir=data_dir)
            worker_result[0] = dq_output
        except Exception as e:
            worker_error[0] = e
        finally:
            _heavy_op_lock.release()
            try:
                if os.path.isfile(job_path):
                    os.remove(job_path)
            except Exception:
                pass

    t = threading.Thread(target=_worker)
    t.start()
    try:
        t.join()
    except BaseException:
        return

    if worker_error[0] is not None:
        st.error(f"Error during data quality check: {worker_error[0]}")
        st.session_state.dq_output = None
        empty_cache()
        return

    st.session_state.dq_output = worker_result[0]
    empty_cache()
    if worker_result[0] is None:
        st.warning(
            "Data quality check skipped: dataset is too small for quality evaluation."
        )


def apply_dq_filter():
    dq_out = st.session_state.get("dq_output")
    if dq_out is None:
        return
    indices = set(getattr(dq_out, "indices_to_remove", None) or [])
    filt = dq_out.splits_filtered
    if isinstance(filt, DatasetDict) and "train" in filt:
        df = filt["train"].to_pandas()
    elif indices and not st.session_state.train_data.empty:
        df = st.session_state.train_data.drop(
            index=[i for i in indices if i < len(st.session_state.train_data)],
            errors="ignore",
        ).reset_index(drop=True)
    else:
        return
    if "text" not in df.columns or "label" not in df.columns:
        return
    cols = ["text", "label"]
    if "anc_label" in df.columns:
        cols.append("anc_label")
    st.session_state.train_data = df[cols].copy()
    ensure_dirs()
    st.session_state.train_data.to_csv(get_train_path(), index=False)
    st.session_state.dq_apply_success = True


def apply_dq_filter_ner():
    dq_out = st.session_state.get("dq_output")
    if dq_out is None:
        return
    indices = set(getattr(dq_out, "indices_to_remove", None) or [])
    filt = dq_out.splits_filtered
    if isinstance(filt, DatasetDict) and "train" in filt:
        train_ds = filt["train"]
        if (
            "text" not in train_ds.column_names
            or "tokens" not in train_ds.column_names
            or "labels" not in train_ds.column_names
        ):
            return
        language = st.session_state.get("language", "en")
        records = []
        for row in train_ds:
            text = row["text"]
            tokens = row["tokens"]
            labels = row["labels"]
            spans = convert_bio_to_spans(text, tokens, labels, language=language)
            records.append({"text": text, "spans": spans})
    elif indices and st.session_state.ner_train_data:
        records = [
            r for i, r in enumerate(st.session_state.ner_train_data) if i not in indices
        ]
    else:
        return
    st.session_state.ner_train_data = records
    ensure_dirs()
    with open(get_ner_train_path(), "w", encoding="utf-8") as f:
        json.dump(st.session_state.ner_train_data, f, ensure_ascii=False)
    st.session_state.dq_apply_success = True


# Training


def _run_training_impl(job_path: str):
    if is_ner():
        _run_ner_training(job_path)
    else:
        _run_text_clf_training(job_path)


def run_training():
    sid = _session_id()
    with _queue_lock:
        if any(s == sid for s, _ in _heavy_op_queue):
            return
        _heavy_op_queue.append((sid, "training"))
    st.session_state.queued_action = "training"
    st.session_state.tab_choice = "Training"
    _save_session_meta()


def _get_llm_client_overrides() -> Optional[Dict[str, Any]]:
    llm_cfg = st.session_state.get("llm_config", {})
    api_key = (llm_cfg.get("api_key") or "").strip()
    base_url = (llm_cfg.get("base_url") or "").strip()
    if not api_key or not base_url:
        return None
    return {
        "LlmClientConfig": {
            "api_key": api_key,
            "base_url": base_url,
            "model_id": (llm_cfg.get("model_id") or "").strip() or None,
        }
    }


def _run_text_clf_training(job_path: str):
    if st.session_state.train_data.empty:
        st.error("Training data is missing")
        return
    save_inputs()

    config = st.session_state.training_config
    ood = OodMethod[config["ood_method"]]
    config_overrides: Dict[str, Any] = {
        "ood_method": ood,
        "language": st.session_state.get("language", "en"),
    }

    if ood not in (OodMethod.NONE, OodMethod.LOGIT):
        config_overrides["threshold_factor"] = config.get("threshold_factor", 1.0)

    if config.get("llm_augmentation", False):
        llm_overrides = _get_llm_client_overrides()
        if llm_overrides is None:
            st.error(
                "LLM Augmentation is enabled but API key and Base URL are not set. "
                "Configure them in the Training Configuration section."
            )
            return
        config_overrides["llm_augmentation"] = {
            "enabled": True,
            "use_domain_analysis": True,
            "config_overrides": llm_overrides,
        }

    if config.get("llm_test_generation", False):
        llm_overrides = _get_llm_client_overrides()
        if llm_overrides is None:
            st.error(
                "Test generation is enabled but API key and Base URL are not set. "
                "Configure them in the Training Configuration section."
            )
            return
        config_overrides["llm_test_generation"] = {
            "enabled": True,
            "num_samples_per_class": 100,
            "use_domain_analysis": True,
            "config_overrides": llm_overrides,
        }

    train_path = get_train_path()
    test_path = get_test_path() if st.session_state.test_data is not None else None
    model_path = os.path.join(_session_model_dir(), f"model_{int(time.time())}")
    results_path = os.path.join(_session_data_dir(), _TRAINING_RESULTS_FILENAME)
    meta_path = _session_meta_path()
    ensure_dirs()

    worker_error = [None]

    worker_error = [None]

    def _worker():
        try:
            pipeline = TextClassificationTrainingPipeline(
                train_path=train_path,
                test_path=test_path,
                config_overrides=config_overrides,
            )
            artifact = pipeline.train()
            method_name = pipeline.training_method_cls.__name__
            pipeline.save(model_path, SaveFormat.TORCH)
            _save_training_results(artifact, method_name, path=results_path)
            _update_meta_model_path(meta_path, model_path)
        except Exception as e:
            worker_error[0] = e
        finally:
            _heavy_op_lock.release()
            try:
                os.remove(job_path)
            except OSError:
                pass

    t = threading.Thread(target=_worker)
    t.start()
    try:
        t.join()
    except BaseException:
        return

    if worker_error[0] is not None:
        empty_cache()
        st.error(f"Error during training: {worker_error[0]}")
        return

    loaded = _load_training_results()
    if loaded is not None:
        st.session_state.training_artifact = loaded[0]
        st.session_state.method_name = loaded[1]
    st.session_state.model_path = model_path
    _save_session_meta()
    empty_cache()
    st.success("Training completed successfully!")


def _run_ner_training(job_path: str):
    if st.session_state.ner_train_data is None:
        st.error("NER training data is missing")
        return
    save_ner_inputs()
    _save_session_meta()
    lang = st.session_state.get("language", "en")
    ner_train_path = get_ner_train_path()
    ner_test_path = (
        get_ner_test_path() if st.session_state.ner_test_data is not None else None
    )
    model_path = os.path.join(_session_model_dir(), f"ner_model_{int(time.time())}")
    results_path = os.path.join(_session_data_dir(), _TRAINING_RESULTS_FILENAME)
    meta_path = _session_meta_path()
    ensure_dirs()

    worker_error = [None]

    def _worker():
        try:
            pipeline = TokenClassificationTrainingPipeline(
                train_path=ner_train_path,
                test_path=ner_test_path,
                config_overrides={"language": lang},
            )
            artifact = pipeline.train()
            method_name = pipeline.training_method_cls.__name__
            pipeline.save(model_path, SaveFormat.TORCH)
            _save_training_results(artifact, method_name, path=results_path)
            _update_meta_model_path(meta_path, model_path)
        except Exception as e:
            worker_error[0] = e
        finally:
            _heavy_op_lock.release()
            try:
                os.remove(job_path)
            except OSError:
                pass

    t = threading.Thread(target=_worker)
    t.start()
    try:
        t.join()
    except BaseException:
        return

    if worker_error[0] is not None:
        empty_cache()
        st.error(f"Error during training: {worker_error[0]}")
        return

    loaded = _load_training_results()
    if loaded is not None:
        st.session_state.training_artifact = loaded[0]
        st.session_state.method_name = loaded[1]
    st.session_state.model_path = model_path
    _save_session_meta()
    empty_cache()
    st.success("Training completed successfully!")


# Inference


def predict_texts(texts: List[str]):
    if st.session_state.model_path is None:
        st.error("Trained model is missing")
        return None
    if is_ner():
        return _predict_ner(texts)
    return _predict_text_clf(texts)


def _predict_text_clf(texts: List[str]):
    manager = TextClassificationInferenceManager(st.session_state.model_path)
    outputs = manager.predict(texts, batch_size=32, return_all_hypotheses=False)
    rows = []
    for output in outputs:
        best = output.most_probable
        rows.append({"text": output.text, "label": best.label, "score": best.score})
    return pd.DataFrame(rows)


def _predict_ner(texts: List[str]):
    manager = TokenClassificationInferenceManager(st.session_state.model_path)
    outputs = manager.predict(texts, batch_size=32)
    rows = []
    for output in outputs:
        for entity in output.labels:
            rows.append(
                {
                    "text": output.text[:60] + "..."
                    if len(output.text) > 60
                    else output.text,
                    "entity": entity.text,
                    "label": entity.label,
                    "start": entity.start,
                    "end": entity.end,
                    "score": round(entity.score, 4),
                }
            )
    if not rows:
        return pd.DataFrame(
            columns=["text", "entity", "label", "start", "end", "score"]
        )
    return pd.DataFrame(rows)


# Cache


def clear_cache():
    data_dir = _session_data_dir()
    for name in (
        _TRAINING_RESULTS_FILENAME,
        _DQ_RESULTS_FILENAME,
        _DQ_REPORT_FILENAME,
    ):
        path = os.path.join(data_dir, name)
        if os.path.isfile(path):
            try:
                os.remove(path)
            except Exception:
                pass
    for key in list(st.session_state.keys()):
        del st.session_state[key]
    st.cache_data.clear()
    st.cache_resource.clear()
    empty_cache()


# Display helpers


def display_training_results(tab, artifact, method_name: str):
    tab.subheader("Training Results")
    tab.write(f"**Method:** {method_name}")

    if artifact.labels:
        tab.write(f"**Number of classes:** {len(artifact.labels)}")

    if artifact.test_metrics is not None:
        _display_test_metrics(tab, artifact)
    else:
        tab.info("Test dataset is missing", icon="ℹ️")


def _display_test_metrics(tab, artifact):
    tab.subheader("Test Set Metrics")
    test_report = pd.DataFrame(artifact.test_metrics.classification_report).T

    col1, col2 = tab.columns(2)
    if "macro avg" in test_report.index and "f1-score" in test_report.columns:
        col1.metric(
            "F1-Score (Macro)", f"{test_report.loc['macro avg', 'f1-score']:.4f}"
        )
    if "weighted avg" in test_report.index and "f1-score" in test_report.columns:
        col2.metric(
            "F1-Score (Weighted)", f"{test_report.loc['weighted avg', 'f1-score']:.4f}"
        )

    class_labels = [label for label in test_report.index if label not in AGGREGATE_ROWS]

    if (
        artifact.test_metrics.confusion_matrix is not None
        and artifact.test_metrics.confusion_matrix.size > 0
    ):
        _display_confusion_matrix(
            tab, artifact.test_metrics.confusion_matrix, class_labels, artifact.labels
        )


def _display_confusion_matrix(
    tab, cm, class_labels: List[str], fallback_labels: Optional[List[str]]
):
    try:
        labels = class_labels if class_labels else fallback_labels
        if not labels:
            tab.warning("Cannot display confusion matrix: no labels available")
            return
        if len(labels) != cm.shape[0]:
            if fallback_labels and len(fallback_labels) == cm.shape[0]:
                labels = fallback_labels
            else:
                return

        tab.subheader("Confusion Matrix")
        fig_cm = px.imshow(
            cm,
            labels=dict(x="Predicted", y="Actual", color="Count"),
            x=labels,
            y=labels,
            aspect="auto",
            color_continuous_scale="Blues",
        )
        cm_height = max(600, len(labels) * 25)
        fig_cm.update_layout(height=cm_height)
        tab.plotly_chart(fig_cm, use_container_width=True)
    except Exception as e:
        tab.warning(f"Could not display confusion matrix: {str(e)}")


# UI

init_state()


@st.fragment(run_every=timedelta(seconds=_QUEUE_REFRESH_SECONDS))
def _heavy_op_queue_processor():
    sid = _session_id()
    if st.session_state.get("training_in_progress"):
        if not os.path.isfile(_job_running_path()):
            st.session_state.training_in_progress = False
            meta = _load_session_meta()
            meta_model = meta.get("model_path") if meta else None
            if meta_model and os.path.isdir(meta_model):
                st.session_state.model_path = meta_model
            elif st.session_state.get("model_path") is None:
                st.session_state.model_path = _find_latest_model_dir()
            if st.session_state.get("training_artifact") is None:
                loaded = _load_training_results()
                if loaded is not None:
                    st.session_state.training_artifact = loaded[0]
                    st.session_state.method_name = loaded[1]
            if st.session_state.get("dq_output") is None:
                loaded_dq = _load_dq_results()
                if loaded_dq is not None:
                    st.session_state.dq_output = loaded_dq
            st.rerun(scope="app")
    with _queue_lock:
        if not _heavy_op_queue or _heavy_op_queue[0][0] != sid:
            return
        item = _heavy_op_queue.popleft()
    action = item[1]
    if not _heavy_op_lock.acquire(blocking=False):
        with _queue_lock:
            _heavy_op_queue.appendleft(item)
        return
    st.session_state.training_in_progress = True
    st.session_state.queued_action = None
    st.session_state.tab_choice = "Training" if action == "training" else "Data Quality"
    ensure_dirs()
    job_path = _job_running_path()
    try:
        with open(job_path, "w", encoding="utf-8") as f:
            json.dump({"action": action}, f)
    except Exception:
        pass
    if action == "training":
        _run_training_impl(job_path)
    else:
        _run_dq_impl(job_path)
    try:
        st.session_state.training_in_progress = False
    except Exception:
        pass
    try:
        _save_session_meta()
        st.rerun(scope="app")
    except Exception:
        pass


_heavy_op_queue_processor()

st.title("OpenAutoNLU Pipeline")

col_task, col_lang = st.columns(2)
with col_task:
    st.selectbox(
        "Task Type",
        options=[TASK_TEXT_CLF, TASK_NER],
        key="task_type",
        on_change=on_task_type_change,
    )
with col_lang:
    st.selectbox(
        "Language",
        options=["ru", "en"],
        format_func=lambda x: "Russian" if x == "ru" else "English",
        key="language",
        on_change=_save_session_meta,
        help="Affects default model (ruBert / BERT) and tokenization.",
    )

_TAB_OPTIONS = ["Data Input", "Data Quality", "Training", "Inference", "Clear Cache"]
if "tab_choice" not in st.session_state:
    st.session_state.tab_choice = "Data Input"
st.radio(
    "Tab",
    options=_TAB_OPTIONS,
    key="tab_choice",
    horizontal=True,
    label_visibility="collapsed",
    on_change=_save_session_meta,
)

data_input = st.container()
data_quality_tab = st.container()
training_tab = st.container()
inference_tab = st.container()
cache_tab = st.container()


@st.fragment(run_every=timedelta(seconds=_QUEUE_REFRESH_SECONDS))
def _data_quality_tab_content():
    if st.session_state.get("training_in_progress"):
        st.info(
            "Data quality check in progress... The page will update when it finishes "
            f"(refresh every {_QUEUE_REFRESH_SECONDS} s)."
        )
    if is_ner():
        if st.session_state.ner_train_data is not None:
            st.write(f"Train data: **{len(st.session_state.ner_train_data)} records**")
        else:
            st.info("Upload NER training data first", icon="ℹ️")
    else:
        st.dataframe(st.session_state.train_data, width=600)

    if st.session_state.get("queued_action") == "dq":
        pos = _current_queue_position()
        pos_text = f" (position {pos})" if pos is not None else ""
        st.info(
            f"You are in the queue{pos_text}. "
            f"The queue is checked every {_QUEUE_REFRESH_SECONDS} s; the status box below will update. "
            "When it's your turn, the check will start automatically"
        )
    st.button("Check Data Quality", key="dq", on_click=run_dq)

    if st.session_state.dq_output is not None:
        dq = st.session_state.dq_output
        num_candidates = len(list(dq.indices_to_remove))
        st.caption(
            "Training uses the current training data. "
            "Data quality check only marks candidates for removal. "
            'To train on cleaned data, click "Apply cleanup".'
        )
        if num_candidates == 0:
            st.success("Data quality check completed. No problematic samples found!")
        else:
            st.write(f"Candidates for removal: {num_candidates}")
            if dq.aggregated_report_path and os.path.exists(dq.aggregated_report_path):
                report_df = pd.read_csv(dq.aggregated_report_path)
                st.dataframe(report_df)
            else:
                st.info("Detailed report is being generated...")

            if not is_ner():
                st.button("Apply cleanup", key="apply_dq", on_click=apply_dq_filter)
            else:
                st.button(
                    "Apply cleanup",
                    key="apply_dq_ner",
                    on_click=apply_dq_filter_ner,
                )
        if st.session_state.pop("dq_apply_success", False):
            st.success("Cleaned data applied. Training will use it from now on.")


def _training_tab_content():
    if st.session_state.get("training_in_progress"):
        with st.status("Training in progress...", state="running", expanded=True):
            st.info(
                "Model is training. The page will update when it finishes (refresh every "
                f"{_QUEUE_REFRESH_SECONDS} s)."
            )
    if st.session_state.get("queued_action") == "training":
        pos = _current_queue_position()
        pos_text = f" (position {pos})" if pos is not None else ""
        st.info(
            f"You are in the queue{pos_text}. "
            f"The queue is checked every {_QUEUE_REFRESH_SECONDS} s; the status box below will update. "
            "When it's your turn, training will start automatically "
        )
    if not is_ner():
        st.session_state["_training_config_expanded"] = True
        with st.expander(
            "⚙️ Training Configuration",
            expanded=st.session_state["_training_config_expanded"],
        ):
            ood_method = st.selectbox(
                "OOD Detection Method",
                options=[
                    "AUTO",
                    "NONE",
                    "LOGIT",
                    "MARGINAL_MAHALANOBIS_OOD",
                    "MSP_OOD",
                ],
                index=[
                    "AUTO",
                    "NONE",
                    "LOGIT",
                    "MARGINAL_MAHALANOBIS_OOD",
                    "MSP_OOD",
                ].index(st.session_state.training_config.get("ood_method", "AUTO")),
                help="Method for out-of-distribution detection. AUTO selects automatically based on dataset size.",
            )

            show_threshold = ood_method not in ("NONE", "LOGIT")
            threshold_factor = st.session_state.training_config.get(
                "threshold_factor", 1.0
            )
            if show_threshold:
                threshold_factor = st.slider(
                    "OOD Threshold Factor",
                    min_value=0.0,
                    max_value=10.0,
                    value=float(threshold_factor),
                    step=0.01,
                    help="Multiplier for the OOD detection threshold. "
                    "Higher values = fewer samples detected as OOD (more conservative). "
                    "Default: 1.0",
                )

            llm_augmentation = st.checkbox(
                "LLM Augmentation",
                value=st.session_state.training_config.get("llm_augmentation", False),
                help="Adds synthetic examples for low-resource classes: classes with fewer than 81 examples "
                "are augmented up to 81 per class via LLM. Requires API key and Base URL below.",
            )
            llm_test_generation = st.checkbox(
                "Test generation",
                value=st.session_state.training_config.get(
                    "llm_test_generation", False
                ),
                help="When test set is not uploaded, generates 100 examples per class via LLM. Uses the same API key and Base URL.",
            )

            if "llm_config" not in st.session_state:
                st.session_state.llm_config = _default_llm_config()
            llm_cfg = st.session_state.llm_config

            if llm_augmentation or llm_test_generation:
                st.caption(
                    "LLM configuration (required when augmentation or test generation is enabled). "
                    "Augmentation: classes with < 81 examples are augmented up to 81 per class. "
                    "Test generation: 100 examples per class when test set is not provided."
                )
                llm_api_key = st.text_input(
                    "LLM API Key",
                    value=llm_cfg.get("api_key", ""),
                    type="password",
                    key="llm_api_key",
                    help="API key for the LLM service.",
                )
                llm_base_url = st.text_input(
                    "LLM Base URL",
                    value=llm_cfg.get("base_url", ""),
                    key="llm_base_url",
                    help="Base URL of the LLM API (e.g. https://api.openai.com/v1/). ",
                )
                llm_model_id = st.text_input(
                    "LLM Model ID",
                    value=llm_cfg.get("model_id", ""),
                    key="llm_model_id",
                    help="Model name/ID (e.g. gpt-4). ",
                )
                st.session_state.llm_config = {
                    "api_key": llm_api_key,
                    "base_url": llm_base_url.strip(),
                    "model_id": llm_model_id.strip(),
                }

            st.session_state.training_config = {
                "ood_method": ood_method,
                "threshold_factor": threshold_factor,
                "llm_augmentation": llm_augmentation,
                "llm_test_generation": llm_test_generation,
            }

    st.button("Start Training", key="training", on_click=run_training)
    if st.session_state.training_artifact is not None:
        st.success("Training is complete. You can proceed to the **Inference** tab.")
        artifact = st.session_state.training_artifact
        method_name = st.session_state.method_name or "unknown"
        display_training_results(st, artifact, method_name)


if st.session_state.tab_choice == "Data Input":
    with data_input:
        if is_ner():
            data_input.info(
                "JSON file: list of objects with fields `text` and `spans`.\n"
                "Each span: `{label, start, end}`",
                icon="ℹ️",
            )
            data_input.markdown(
                "<p style='font-weight: 700; font-size: 1.15em;'>Train data</p>",
                unsafe_allow_html=True,
            )
            data_input.file_uploader(
                "",
                type="json",
                key="ner_train_file",
                on_change=add_ner_train_file,
                label_visibility="collapsed",
            )
            data_input.markdown(
                "<p style='font-weight: 700; font-size: 1.15em;'>Test data(optional)</p>",
                unsafe_allow_html=True,
            )
            data_input.file_uploader(
                "",
                type="json",
                key="ner_test_file",
                on_change=add_ner_test_file,
                label_visibility="collapsed",
            )
            _dl_expander_ner = data_input.expander(
                "Download data (example format)", expanded=False
            )
            with _dl_expander_ner:
                dcol1, dcol2 = st.columns(2)
                with dcol1:
                    st.download_button(
                        "Download train.json",
                        data=SAMPLE_TRAIN_NER.read_bytes(),
                        file_name="train.json",
                        mime="application/json",
                        key="download_ner_train",
                    )
                with dcol2:
                    st.download_button(
                        "Download test.json",
                        data=SAMPLE_TEST_NER.read_bytes(),
                        file_name="test.json",
                        mime="application/json",
                        key="download_ner_test",
                    )
            _train_preview_ner = data_input.expander(
                "Train data: preview", expanded=False
            )
            with _train_preview_ner:
                if st.session_state.ner_train_data:
                    summary = ner_data_summary(st.session_state.ner_train_data)
                    st.caption(f"**{len(st.session_state.ner_train_data)} records**")
                    st.dataframe(summary, use_container_width=True)
                else:
                    st.caption("No train data loaded.")
            _test_preview_ner = data_input.expander(
                "Test data: preview", expanded=False
            )
            with _test_preview_ner:
                if st.session_state.ner_test_data:
                    summary_test = ner_data_summary(st.session_state.ner_test_data)
                    st.caption(f"**{len(st.session_state.ner_test_data)} records**")
                    st.dataframe(summary_test, use_container_width=True)
                else:
                    st.caption("No test data loaded")
        else:
            data_input.info(
                "CSV file should contain columns: text, label (optionally anc_label).",
                icon="ℹ️",
            )
            data_input.markdown(
                "<p style='font-weight: 700; font-size: 1.15em;'>Train data</p>",
                unsafe_allow_html=True,
            )
            data_input.file_uploader(
                "",
                type="csv",
                key="train_file",
                on_change=add_train_file,
                label_visibility="collapsed",
            )
            data_input.markdown(
                "<p style='font-weight: 700; font-size: 1.15em;'>Test data(optional)</p>",
                unsafe_allow_html=True,
            )
            data_input.file_uploader(
                "",
                type="csv",
                key="test_file",
                on_change=add_test_file,
                label_visibility="collapsed",
            )
            _dl_expander_csv = data_input.expander(
                "Download data (example format)", expanded=False
            )
            with _dl_expander_csv:
                dcol1, dcol2 = st.columns(2)
                with dcol1:
                    st.download_button(
                        "Download train.csv",
                        data=SAMPLE_TRAIN_TEXT.read_bytes(),
                        file_name="train.csv",
                        mime="text/csv",
                        key="download_train",
                    )
                with dcol2:
                    st.download_button(
                        "Download test.csv",
                        data=SAMPLE_TEST_TEXT.read_bytes(),
                        file_name="test.csv",
                        mime="text/csv",
                        key="download_test",
                    )
            _train_preview = data_input.expander("Train data: preview", expanded=False)
            with _train_preview:
                if not st.session_state.train_data.empty:
                    st.caption(f"**{len(st.session_state.train_data)} rows**")
                    st.dataframe(st.session_state.train_data, use_container_width=True)
                else:
                    st.caption("No train data loaded.")
            _test_preview = data_input.expander("Test data: preview", expanded=False)
            with _test_preview:
                if (
                    st.session_state.test_data is not None
                    and not st.session_state.test_data.empty
                ):
                    st.caption(f"**{len(st.session_state.test_data)} rows**")
                    st.dataframe(st.session_state.test_data, use_container_width=True)
                else:
                    st.caption("No test data loaded")

elif st.session_state.tab_choice == "Data Quality":
    with data_quality_tab:
        _data_quality_tab_content()

elif st.session_state.tab_choice == "Training":
    with training_tab:
        _training_tab_content()

elif st.session_state.tab_choice == "Inference":
    with inference_tab:
        inference_tab.text_area(
            "Text for classification"
            if not is_ner()
            else "Text for entity recognition",
            key="single_text",
        )
        if inference_tab.button("Predict", key="predict_one"):
            if st.session_state.single_text:
                pred_df = predict_texts([st.session_state.single_text])
                if pred_df is not None:
                    inference_tab.dataframe(pred_df, use_container_width=True)
            else:
                inference_tab.error("Please enter text")

        inference_tab.info(
            f"CSV file should contain column: text. Max {MAX_INFERENCE_BATCH_SIZE} rows per file.",
            icon="ℹ️",
        )
        inference_tab.file_uploader(
            "Batch data",
            type="csv",
            key="inference_file",
            on_change=lambda: st.session_state.update(batch_predictions=None),
        )
        if inference_tab.button("Predict batch", key="predict_batch"):
            if st.session_state.inference_file is not None:
                df = pd.read_csv(st.session_state.inference_file)
                if "text" not in df.columns:
                    inference_tab.error("wrong columns, expected: text")
                elif len(df) > MAX_INFERENCE_BATCH_SIZE:
                    inference_tab.error(
                        f"Batch size {len(df)} exceeds maximum allowed ({MAX_INFERENCE_BATCH_SIZE}). "
                        "Use a smaller file or split the data."
                    )
                else:
                    with st.spinner("Running batch prediction..."):
                        st.session_state.batch_predictions = predict_texts(
                            df["text"].tolist()
                        )
            else:
                inference_tab.error("Please upload a file first")

        if st.session_state.batch_predictions is not None:
            inference_tab.dataframe(
                st.session_state.batch_predictions, use_container_width=True
            )

elif st.session_state.tab_choice == "Clear Cache":
    with cache_tab:
        cache_tab.button("Clear Cache", key="clear_cache", on_click=clear_cache)



from .anc_set_fit import AncSetFitMethod
from .anc_set_fit_ood import AncSetFitOOD
from .data_types import SaveFormat
from .finetuner import Finetuner, FinetunerInferenceManager
from .finetuner_ood import FinetunerWithOOD
from .finetuning_training_args import (
    FinetuningTrainingArguments,
    OptimizableHyperParameter,
    OptunaSuggestionDtype,
)
from .method import Method
from .set_fit import SetFitInferenceManager, SetFitMethod
from .set_fit_ood import SetFitOOD
from .token_classification_finetuner import (
    TokenClassificationFinetuner,
    TokenClassificationInferenceManager,
)

__all__ = [
    "AncSetFitMethod",
    "Finetuner",
    "FinetuningTrainingArguments",
    "Method",
    "SetFitMethod",
    "TokenClassificationFinetuner",
    "OptimizableHyperParameter",
    "OptunaSuggestionDtype",
    "FinetunerInferenceManager",
    "SetFitInferenceManager",
    "TokenClassificationInferenceManager",
    "SaveFormat",
    "FinetunerWithOOD",
    "SetFitOOD",
    "AncSetFitOOD",
]



import logging
from dataclasses import replace
from typing import Callable, Dict, Tuple, Optional

import numpy as np
import optuna
from transformers import EarlyStoppingCallback, Trainer, TrainingArguments
from transformers.trainer_utils import HPSearchBackend, IntervalStrategy

from ..constants import HPO_STAGE

from .finetuning_training_args import OptimizableHyperParameter, OptunaSuggestionDtype

log = logging.getLogger(__name__)


class EarlyStoppingCallbackPatched(EarlyStoppingCallback):
    """
    Patched version of the EarlyStoppingCallback to handle specific requirements.
    """

    def on_train_begin(self, args, state, control, **kwargs):
        """
        Method called at the beginning of training.

        Args:
            args: TrainingArguments object.
            state: TrainerState object.
            control: TrainerControl object.
            **kwargs: Additional keyword arguments.
        """
        assert (
            args.metric_for_best_model is not None
        ), "EarlyStoppingCallback requires metric_for_best_model is defined"
        assert (
            args.evaluation_strategy != IntervalStrategy.NO
        ), "EarlyStoppingCallback requires IntervalStrategy of steps or epoch"
        self.early_stopping_patience_counter = 0

    def on_evaluate(self, args, state, control, metrics, **kwargs):
        """
        Method called during evaluation.

        Args:
            args: TrainingArguments object.
            state: TrainerState object.
            control: TrainerControl object.
            metrics: Dictionary of evaluation metrics.
            **kwargs: Additional keyword arguments.
        """
        metric_to_check = args.metric_for_best_model
        if not metric_to_check.startswith("eval_"):
            metric_to_check = f"eval_{metric_to_check}"
        metric_value = metrics.get(metric_to_check)

        if metric_value is None:
            log.warning(
                f"early stopping required metric_for_best_model, but did not find {metric_to_check} so early stopping"
                " is disabled"
            )
            return

        self.check_metric_value(args, state, control, metric_value)

        if self.early_stopping_patience_counter >= self.early_stopping_patience:
            control.should_training_stop = True

        operator = np.greater if args.greater_is_better else np.less
        if state.best_metric is None or operator(metric_value, state.best_metric):
            state.best_metric = metric_value


class ImprovementActivatedEarlyStopping(EarlyStoppingCallback):
    """Early stopping callback that activates only after initial improvements.

    Attributes:
        early_stopping_patience: Number of evaluations without improvement
            before stopping (after activation).
        early_stopping_threshold: Minimum change to qualify as improvement.
        improvement_threshold: Number of improvements required to activate
            early stopping.
    """

    def __init__(
        self,
        early_stopping_patience: int = 3,
        early_stopping_threshold: Optional[float] = 0.0,
        improvement_threshold: Optional[int] = 3,
    ):
        super().__init__(early_stopping_patience, early_stopping_threshold)
        self.early_stopping_patience = early_stopping_patience
        self.improvement_threshold = improvement_threshold
        self.best_metric = None
        self.patience_counter = 0
        self.improvement_counter = 0
        self.early_stopping_active = False

    def on_evaluate(self, args, state, control, metrics=None, **kwargs):
        if metrics is None:
            return

        # Replace 'eval_loss' with your metric name if different
        current_metric = metrics.get("eval_loss")

        if self.best_metric is None:
            # First evaluation
            self.best_metric = current_metric
            return

        if (
            current_metric < self.best_metric
        ):  # Metric improved (use > if lower is better)
            self.improvement_counter += 1
            self.best_metric = current_metric
            self.patience_counter = 0  # Reset patience counter on improvement

            # Activate early stopping after threshold improvements
            if (
                not self.early_stopping_active
                and self.improvement_counter >= self.improvement_threshold
            ):
                print(
                    f"Early stopping activated after {self.improvement_threshold} improvements"
                )
                self.early_stopping_active = True
        else:
            # Metric didn't improve
            if self.early_stopping_active:
                self.patience_counter += 1
                print(
                    f"Early stopping patience: {self.patience_counter}/{self.early_stopping_patience}"
                )

                if self.patience_counter >= self.early_stopping_patience:
                    print("Triggering early stopping")
                    control.should_training_stop = True


class FinetuningHyperparameterSearchMixin:
    """Mixin providing Optuna-based hyperparameter optimization for finetuning.

    Adds hyperparameter search capabilities to transformer training classes.
    Uses Optuna with TPE (Tree-structured Parzen Estimator) sampler for
    efficient hyperparameter exploration.

    Usage:
        Inherit from this mixin and call `optimize_hyperparameters()` with a
        configured Trainer and hyperparameter search space.

    Attributes:
        hyp_search_done: Flag indicating if HPO has been completed.
    """

    @staticmethod
    def _prepare_hp_space(
        **hp_space_config: Dict[str, OptimizableHyperParameter],
    ) -> callable:
        """
        Prepares the hyperparameter space for Optuna.

        Args:
            **hp_space_config: Dictionary of hyperparameter configurations.

        Returns:
            callable: A function that defines the hyperparameter space for Optuna.
        """

        def optuna_hp_space(trial):
            hp_space = {}
            for hp_name, optimizable in hp_space_config.items():
                if optimizable.dtype != OptunaSuggestionDtype.CATEGORICAL:
                    hp_space[hp_name] = getattr(
                        trial, f"suggest_{optimizable.dtype.value}"
                    )(hp_name, *optimizable.range, **optimizable.optuna_kwargs)
                else:
                    hp_space[hp_name] = getattr(
                        trial, f"suggest_{optimizable.dtype.value}"
                    )(hp_name, optimizable.range)
            return hp_space

        return optuna_hp_space

    @staticmethod
    def _prepare_args(args: TrainingArguments) -> Tuple[TrainingArguments, dict]:
        """
        Prepares the training arguments for hyperparameter search.

        Args:
            args (TrainingArguments): The training arguments.

        Returns:
            dict: A dictionary containing the modified arguments and a backup of the original values.
        """
        backup = {
            "load_best_model_at_end": args.load_best_model_at_end,
            "save_strategy": args.save_strategy,
        }
        args.load_best_model_at_end = False
        args.save_strategy = IntervalStrategy.NO
        return args, backup

    @staticmethod
    def _restore_args(args_backup: dict, args: TrainingArguments) -> TrainingArguments:
        """
        Restores the training arguments to their original values.

        Args:
            args_backup (dict): A dictionary containing the backup of the original arguments.
            args (TrainingArguments): The training arguments.

        Returns:
            TrainingArguments: The restored training arguments.
        """
        args = replace(args, **args_backup)
        return args

    @staticmethod
    def _patch_or_add_early_stopping(trainer: Trainer):
        """
        Patches the early stopping callback in the trainer.

        Args:
            trainer (Trainer): The trainer object.

        Returns:
            Trainer: The patched trainer object.
        """
        old_callback = trainer.callback_handler.pop_callback(EarlyStoppingCallback)
        if old_callback:
            trainer.add_callback(
                EarlyStoppingCallbackPatched(
                    old_callback.early_stopping_patience,
                    old_callback.early_stopping_threshold,
                )
            )
        else:
            trainer.add_callback(EarlyStoppingCallbackPatched(5))
        return trainer, old_callback

    @staticmethod
    def _restore_early_stopping_setup(
        trainer: Trainer, old_callback: Optional[EarlyStoppingCallback] = None
    ):
        """
        Restores the original early stopping callback in the trainer.

        Args:
            trainer (Trainer): The trainer object.

        Returns:
            Trainer: The restored trainer object.
        """
        patched_callback = trainer.callback_handler.pop_callback(
            EarlyStoppingCallbackPatched
        )
        if old_callback:
            trainer.add_callback(old_callback)
        elif patched_callback:
            trainer.add_callback(
                EarlyStoppingCallback(
                    patched_callback.early_stopping_patience,
                    patched_callback.early_stopping_threshold,
                )
            )
        return trainer

    def optimize_hyperparameters(
        self,
        trainer: Trainer,
        optimizables: Dict[str, OptimizableHyperParameter],
        num_trials: int,
        progress_reporter: Callable[[str, float], None] = lambda x, y: None,
    ):
        """
        Performs hyperparameter search using Optuna.

        Returns:
            dict: The best hyperparameters found by Optuna.
        """
        assert (
            not self.hyp_search_done
        ), "The hyperparameter search has already been done"
        trainer.hp_space = self._prepare_hp_space(**optimizables)
        trainer.hp_search_backend = HPSearchBackend.OPTUNA
        trainer.compute_objective = lambda x: x["eval_f1"]

        trainer.args, backup = self._prepare_args(trainer.args)
        trainer, early_stopping_callback = self._patch_or_add_early_stopping(trainer)

        def objective(trial):
            trainer.train(trial=trial)
            return trainer.state.best_metric

        study = optuna.create_study(
            direction="maximize",
            sampler=optuna.samplers.TPESampler(seed=trainer.args.seed),
        )
        progress_callback = OptunaProgressReporter(progress_reporter, num_trials)
        study.enqueue_trial(trainer.args.hyp_search_defaults)
        study.optimize(objective, n_trials=num_trials, callbacks=[progress_callback])

        trainer.args = self._restore_args(backup, trainer.args)
        trainer = self._restore_early_stopping_setup(trainer, early_stopping_callback)
        return study.best_trial.params


class OptunaProgressReporter:
    """Optuna callback for reporting HPO progress.

    Reports progress as a fraction of completed trials to a progress reporter
    function. Used to integrate Optuna optimization with the library's
    progress reporting system.

    Attributes:
        progress_reporter: Callable taking (stage_name, progress_fraction).
        n_trials: Total number of trials planned.
        fraction_per_trial: Progress increment per completed trial.
    """

    def __init__(self, progress_reporter: Callable[[str, float], None], n_trials: int):
        self.progress_reporter = progress_reporter
        self.n_trials = n_trials
        self.fraction_per_trial = 1 / n_trials

    def __call__(
        self, study: optuna.study.Study, trial: optuna.trial.FrozenTrial
    ) -> None:
        if trial.state.is_finished():
            self.progress_reporter(
                HPO_STAGE, (trial.number + 1) * self.fraction_per_trial
            )



import json
import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Type

from datasets import Dataset

from .. import methods
from ..constants import (
    METHOD_HUMAN_READABLE_NAME_MAP,
    OOD_METHOD_MAP,
    TRAIN_KEY,
)
from ..data.simple_data_provider import AbstractDataProvider
from ..data_quality.data_types import DatasetEvaluatorOutput
from ..data_quality.dynamic_finetuner import DynamicTuner
from ..data_quality.evaluator import DatasetEvaluator
from ..data_quality.sample_eval import AbstractSampleEvalMethod
from ..data_types import ProgressCallbackPayload
from ..llm_pipelines import analyze_domain_and_labels
from ..methods.data_types import OodMethod
from ..methods.method import Method, SequenceClassificationMethod
from ..progress_reporter import ProgressReporter
from .constants import LEGACY_META_FILE_NAME, META_FILE_NAME


def get_method_from_string(method_name: str):
    class_ = getattr(methods, method_name)
    return class_


class AbstractTrainingPipeline(ABC):
    """Abstract base class for all training pipelines.

    This class defines the interface and common functionality for training
    pipelines that handle data loading, preprocessing, method selection,
    training, and model serialization.

    Attributes:
        data_provider_cls: Class used to load and parse training data.
        data_quality_tuner_cls: Class for dynamic tuning during data quality evaluation.
        data_quality_evaluators_classes: List of sample evaluation method classes.
        training_method_cls: The training method class to use.
        task_type: String identifier for the task type (e.g., "multiclass_classification").
    """

    data_provider_cls: Type[AbstractDataProvider]
    data_quality_tuner_cls: Type[DynamicTuner]
    data_quality_evaluators_classes: List[Type[AbstractSampleEvalMethod]]
    training_method_cls: Type[methods.Method]
    task_type: str

    def __init__(
        self,
        train_path: str,
        test_path: Optional[str] = None,
        on_progress_changed: Callable[[ProgressCallbackPayload], None] = lambda x: None,
        config_overrides: Optional[Dict[str, Any]] = None,
    ):
        """Initialize the training pipeline.

        Args:
            train_path: Path to the training data file (CSV format).
            test_path: Optional path to the test data file. If not provided,
                test metrics will not be computed after training.
            on_progress_changed: Callback function invoked with progress updates
                during training. Receives a ProgressCallbackPayload object.
            config_overrides: Dictionary to override default configurations.
                Supports multiple resolution strategies:

                1. **By method class name**: Use the method class name as key.
                   ``{"SetFitMethod": {"num_iterations": 30}}``

                2. **By config class name**: Use the config class name as key.
                   ``{"SetFitMethodConfig": {"num_iterations": 30}}``

                3. **Direct field names**: Specify config fields directly.
                   ``{"num_iterations": 30, "batch_size": 32}``

                **Common options:**

                - ``ood_method``: OOD detection method. Values: ``OodMethod.AUTO``,
                  ``OodMethod.NONE``, ``OodMethod.LOGIT``, ``OodMethod.MAHALANOBIS``,
                  ``OodMethod.MSP``.

                - ``llm_augmentation``: LLM-based data augmentation settings::

                    {
                        "llm_augmentation": {
                            "enabled": True,
                            "use_domain_analysis": True,
                            "config_overrides": {
                                "LlmClientConfig": {
                                    "api_key": "your-key",
                                    "base_url": "https://api.example.com/v1/"
                                }
                            }
                        }
                    }

                - ``llm_test_generation``: Synthetic test data generation::

                    {
                        "llm_test_generation": {
                            "enabled": True,
                            "num_samples_per_class": 100,
                            "use_domain_analysis": True,
                            "synthetic_test_path": "./synthetic_test.csv"
                        }
                    }
        """
        language = "en"
        if config_overrides is not None and "language" in config_overrides:
            language = config_overrides["language"]
        self.language = language
        self.splits = self.data_provider_cls(
            train_path=train_path, test_path=test_path, language=language
        ).load()
        self.progress_callback = on_progress_changed
        self.config_overrides = config_overrides
        self._domain_desc: Optional[str] = None
        self._label_descriptions: Optional[Dict[str, str]] = None
        self._log = logging.getLogger(self.__class__.__name__)

    def evaluate_training_data(self) -> DatasetEvaluatorOutput:
        """Evaluate training data quality using configured evaluators.

        Runs the data quality pipeline on the training split, using sample
        evaluators to identify problematic samples (e.g., mislabeled, ambiguous,
        or low-confidence samples).

        Returns:
            DatasetEvaluatorOutput containing evaluation results, including
            filtered dataset and per-sample quality scores.
        """
        method_kwargs = self._resolve_config_overrides(self.data_quality_tuner_cls)
        sample_evaluators = [cls() for cls in self.data_quality_evaluators_classes]  # type: ignore
        progress_reporter = ProgressReporter(self.progress_callback)
        progress_reporter.stages = [e.name for e in sample_evaluators]
        tuner = self.data_quality_tuner_cls(
            self.data_quality_tuner_cls.config_cls(**method_kwargs),
        )
        dataset_evaluator = DatasetEvaluator(
            task_type=self.task_type,
            tuner=tuner,
            sample_evaluators=sample_evaluators,
        )
        # TODO: refactor DQ to directly take in only a Dataset, not a DatasetDict
        # As for now it processes only "train" split, but semantically code looks
        # like it processes all splits.
        evaluation_result: DatasetEvaluatorOutput = dataset_evaluator.filter_data(
            self.splits,
            progress_reporter=progress_reporter,
        )
        return evaluation_result

    # TODO: refactor to use from Mixin
    def _resolve_config_overrides(self, method_cls):
        """Resolve configuration overrides for a given method class.

        Extracts relevant configuration parameters from config_overrides
        using multiple resolution strategies:

        1. If method_cls has ``parent_finetuner_cls_name``, look up by that name
        2. Otherwise, look up by ``method_cls.__name__``
        3. If not found, extract fields that match ``method_cls.config_cls`` attributes

        Args:
            method_cls: The method class to resolve configuration for.

        Returns:
            Dictionary of configuration parameters for the method.
        """
        if self.config_overrides is None:
            return {}

        if hasattr(method_cls, "parent_finetuner_cls_name"):
            method_kwargs = self.config_overrides.get(
                method_cls.parent_finetuner_cls_name
            )
        else:
            method_kwargs = self.config_overrides.get(method_cls.__name__)
        if method_kwargs is None:
            method_kwargs = {
                k: v
                for k, v in self.config_overrides.items()
                if hasattr(method_cls.config_cls, k)
            }
        if method_kwargs.get("ood_method") == OodMethod.AUTO:
            method_kwargs.pop("ood_method")
        return method_kwargs

    @abstractmethod
    def diagnose(self) -> Optional[DatasetEvaluatorOutput]:
        """
        Run the DataQuality untility on splits["train"] and return the result.
        """
        ...

    @abstractmethod
    def _synergize_data_and_method(self) -> tuple[Type[Method], Dataset]:
        """
        Analyze the incoming train dataset, verify its validity
        in terms of data quantity. Upsample, or downsample it,
        and finally resolve the final training method.
        """
        ...

    @property
    def frontend_training_method_name(self) -> str:
        """Get the human-readable name of the training method.

        Returns:
            Human-readable method name for display in frontend/UI.
        """
        return METHOD_HUMAN_READABLE_NAME_MAP[self.training_method_cls.__name__]

    def train(self) -> methods.data_types.TrainingArtifactInfo:
        """Train the model using the configured method and data.

        This method:
        1. Calls _synergize_data_and_method to select the appropriate method
           and preprocess data (upsample/downsample as needed)
        2. Resolves configuration overrides for the selected method
        3. Trains the model on the prepared data
        4. Optionally evaluates on test data if available

        Returns:
            TrainingArtifactInfo containing training results, metrics,
            and model artifacts.
        """
        self.training_method_cls, self.training_data = self._synergize_data_and_method()
        print(self.training_method_cls)
        method_kwargs = self._resolve_config_overrides(self.training_method_cls)
        progress_reporter = ProgressReporter(self.progress_callback)
        self.method_manager = self.training_method_cls(
            self.training_method_cls.config_cls(**method_kwargs),
            progress_reporter=progress_reporter,
        )
        training_artifact_info = self.method_manager.train(self.training_data)
        if test_data := self.splits.get("test"):
            if isinstance(self.method_manager, SequenceClassificationMethod):
                training_artifact_info.test_metrics = self.method_manager.test(
                    test_data, inscope_only=False
                )
            else:
                training_artifact_info.test_metrics = self.method_manager.test(
                    test_data
                )
        return training_artifact_info

    def _maybe_analyze_domain(
        self,
        use_domain_analysis: bool,
        config_overrides: Optional[Dict[str, Any]] = None,
    ) -> None:
        if not use_domain_analysis:
            return
        if self._domain_desc is not None and self._label_descriptions is not None:
            return

        train_df = self.splits[TRAIN_KEY].to_pandas()
        try:
            label_counts = train_df.groupby("label").size()
            min_label_count = label_counts.min() if len(label_counts) > 0 else 0
            examples_per_label = min(7, min_label_count) if min_label_count > 0 else 5

            self._domain_desc, self._label_descriptions = analyze_domain_and_labels(
                reference_data=train_df,
                text_column="text",
                label_column="label",
                examples_per_label=examples_per_label,
                config_overrides=config_overrides,
                language=self.language,
            )
        except Exception as exc:  # pylint: disable=broad-except
            self._log.warning(
                "Failed to perform domain analysis: %s", exc, exc_info=True
            )

    def save(self, path_to_package: str, save_format: methods.data_types.SaveFormat):
        """Save the trained model to disk.

        Serializes the model and writes metadata to the specified directory.
        The metadata file contains information about the method class and
        save format for later reconstruction during inference.

        Args:
            path_to_package: Directory path where the model will be saved.
            save_format: Format for saving (SaveFormat.TORCH or SaveFormat.ONNX).
        """
        self.method_manager.save(path_to_package, save_format)
        with open(path_to_package + "/meta.json", "w") as f:
            json.dump(
                {
                    "method_class": self.training_method_cls.__name__,
                    "type": save_format.value,
                },
                f,
            )


class AbstractInferencePipeline(ABC):
    """Abstract base class for inference pipelines.

    Loads a trained model from disk and provides inference capabilities.
    Automatically detects the model format (Torch or ONNX) and loads
    the appropriate inference manager.
    """

    def __init__(self, path_to_package: str) -> None:
        """Initialize the inference pipeline from a saved model.

        Args:
            path_to_package: Path to the directory containing the saved model
                and metadata file (meta.json).
        """
        path_to_package_obj = Path(path_to_package)
        if (path_to_package_obj / LEGACY_META_FILE_NAME).exists():
            with (path_to_package_obj / LEGACY_META_FILE_NAME).open() as f:
                meta = {
                    "method_class": OOD_METHOD_MAP[f.read().strip()],
                    "type": "torch",
                }
        else:
            with open(path_to_package_obj / META_FILE_NAME) as f:
                meta = json.load(f)
        producer_cls = get_method_from_string(meta["method_class"])
        if (
            methods.data_types.SaveFormat(meta["type"])
            == methods.data_types.SaveFormat.TORCH
        ):
            self.inferer = producer_cls.load(path_to_package)
        else:
            self.inferer = producer_cls.onnx_inference_manager(path_to_package)



import json
import os
from abc import ABC, abstractmethod
from bisect import bisect_left
from typing import List

import numpy as np
import onnxruntime
import torch
from accelerate import find_executable_batch_size
from transformers import AutoTokenizer

from ..methods.constants import BATCH_SIZE
from ..methods.data_types import (
    TextClassifierOutput,
    TextLabel,
    TokenClassifierOutput,
)
from ..methods.method import SequenceClassificationMethod
from ..methods.utils import batch as batcher

_PADDING_BUCKETS: tuple[int, ...] = (4, 8, 16, 32, 64, 128, 256, 512)


def _compute_bucket_pad_length(seq_length: int, model_max_length: int) -> int:
    index = bisect_left(_PADDING_BUCKETS, seq_length)
    if index < len(_PADDING_BUCKETS):
        return min(_PADDING_BUCKETS[index], model_max_length)
    return model_max_length


def _pad_inputs_to_bucket(
    inputs: dict[str, np.ndarray],
    model_max_length: int,
    pad_token_id: int,
) -> dict[str, np.ndarray]:
    seq_length = inputs["input_ids"].shape[1]
    bucket_length = _compute_bucket_pad_length(seq_length, model_max_length)
    if seq_length >= bucket_length:
        return inputs
    pad_width = bucket_length - seq_length
    padded: dict[str, np.ndarray] = {}
    for key, value in inputs.items():
        pad_value = pad_token_id if key == "input_ids" else 0
        padded[key] = np.pad(value, ((0, 0), (0, pad_width)), constant_values=pad_value)
    return padded


class AbstractONNXInferenceManager(ABC):
    """Abstract base class for ONNX model inference.

    Provides common infrastructure for loading and running ONNX models
    with automatic hardware acceleration detection. Subclasses implement
    task-specific prediction logic.

    Attributes:
        ort_provider: ONNX Runtime execution provider (CPU, CUDA, or CoreML).
    """

    @abstractmethod
    def __init__(self, path_to_package: str) -> None:
        """Initialize the inference manager.

        Automatically detects available hardware accelerators and selects
        the appropriate ONNX Runtime execution provider.

        Args:
            path_to_package: Path to directory containing model.onnx and
                associated files (tokenizer, label mapping).
        """
        self.ort_provider = "CPUExecutionProvider"
        if torch.cuda.is_available():
            self.ort_provider = "CUDAExecutionProvider"
        elif torch.backends.mps.is_available():
            self.ort_provider = "CoreMLExecutionProvider"

    @abstractmethod
    def predict(
        self,
        texts: List[str],
        batch_size: int = BATCH_SIZE,
        auto_find_batch_size: bool = False,
    ) -> list[TextClassifierOutput] | list[TokenClassifierOutput]:
        """Run inference on input texts.

        Args:
            texts: List of input texts to classify.
            batch_size: Batch size for inference.
            auto_find_batch_size: If True, automatically find the largest
                batch size that fits in memory.

        Returns:
            List of classification outputs.
        """
        ...


class GeneralSequenceClassifierInferenceManager(AbstractONNXInferenceManager):
    """ONNX inference manager for sequence classification models.

    Loads an ONNX-exported text classification model and provides batched
    inference with optional automatic batch size detection. Supports returning
    either the top prediction or all class hypotheses with scores.

    The model package directory must contain:
        - model.onnx: The exported ONNX model
        - tokenizer files: HuggingFace tokenizer configuration
        - label_mapping.json: Mapping of label IDs to label names

    Attributes:
        tokenizer: HuggingFace tokenizer for input preprocessing.
        session: ONNX Runtime inference session.
        id2label: Mapping from numeric IDs to label strings.

    Example:
        >>> manager = GeneralSequenceClassifierInferenceManager("./model_package")
        >>> outputs = manager.predict(["Hello world", "Test text"])
        >>> print(outputs[0].most_probable.label)
    """

    def __init__(self, path_to_package: str) -> None:
        super().__init__(path_to_package)
        self.tokenizer = AutoTokenizer.from_pretrained(path_to_package)
        self.session = onnxruntime.InferenceSession(
            os.path.join(path_to_package, "model.onnx"),
            providers=[self.ort_provider, "CPUExecutionProvider"],
        )
        with open(os.path.join(path_to_package, "label_mapping.json"), "r") as f:
            config = json.load(f)
        self.id2label = {int(key): value for key, value in config["id2label"].items()}

    def predict(
        self,
        texts: List[str],
        batch_size: int = BATCH_SIZE,
        auto_find_batch_size: bool = False,
        return_all_hypotheses: bool = False,
    ) -> list[TextClassifierOutput]:
        """Run classification inference on input texts.

        Args:
            texts: List of input texts to classify.
            batch_size: Batch size for inference.
            auto_find_batch_size: If True, automatically find the largest
                batch size that fits in memory.
            return_all_hypotheses: If True, return scores for all classes.
                If False, return only the top prediction.

        Returns:
            List of TextClassifierOutput with predictions for each input.
        """
        if auto_find_batch_size:
            logits = find_executable_batch_size(
                self._get_logits, starting_batch_size=batch_size
            )(texts)  # type: ignore
        else:
            logits = self._get_logits(batch_size, texts)
        if return_all_hypotheses:
            hypotheses_batch = [[] for _ in range(logits.shape[0])]
            for text_id in range(logits.shape[0]):
                for label_id in range(logits.shape[1]):
                    hypotheses_batch[text_id].append(
                        {
                            "label": self.id2label[label_id],
                            "score": float(logits[text_id, label_id]),
                        }
                    )
            hypotheses_batch = [
                list(sorted(hypotheses, key=lambda x: x["score"], reverse=True))
                for hypotheses in hypotheses_batch
            ]
            packaged = []
            for text, output in zip(texts, hypotheses_batch):
                label_hypotheses = [
                    TextLabel(dict_label["label"], dict_label["score"])
                    for dict_label in output
                ]
                packaged.append(
                    TextClassifierOutput(text=text, hypotheses=label_hypotheses)
                )
            return packaged

        all_scores = SequenceClassificationMethod._extract_scores(logits)
        all_labels = SequenceClassificationMethod._extract_labels(logits, self.id2label)
        return [
            TextClassifierOutput(text=text, hypotheses=[TextLabel(label, score)])
            for label, score, text in zip(all_labels, all_scores, texts)
        ]

    def _get_logits(
        self,
        batch_size: int,
        texts: List[str],
    ) -> np.ndarray:
        total = len(texts)
        if total == 0:
            return np.empty((0, 0), dtype=np.float32)

        logits_buffer = None
        offset = 0
        for text_batch in batcher(texts, batch_size):
            inputs = self.tokenizer(
                text_batch,
                padding=True,
                truncation=True,
                max_length=self.tokenizer.model_max_length,
                return_attention_mask=True,
                return_token_type_ids=True,
                return_tensors="np",
            ).data
            inputs = _pad_inputs_to_bucket(
                inputs, self.tokenizer.model_max_length, self.tokenizer.pad_token_id
            )
            logits = self.session.run(["logits"], inputs)[0]
            if logits_buffer is None:
                logits_buffer = np.empty((total, logits.shape[1]), dtype=logits.dtype)
            batch_size_actual = logits.shape[0]
            logits_buffer[offset : offset + batch_size_actual] = logits
            offset += batch_size_actual

        return logits_buffer if logits_buffer is not None else np.empty((0, 0))



# Configuration file for the Sphinx documentation builder.

import sys
import tomllib
from datetime import datetime
from pathlib import Path

ROOT = Path(__file__).resolve().parents[2]
sys.path.insert(0, str(ROOT))

# -- Project information -----------------------------------------------------
with open(ROOT / "pyproject.toml", "rb") as f:
    pyproject = tomllib.load(f)

project = pyproject["project"]["name"]
copyright = f"{datetime.now().year}, MWS AI"
author = "MWS AI Team"
version = pyproject["project"]["version"]
release = version

# -- General configuration ---------------------------------------------------
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "autoapi.extension",
    "sphinx.ext.viewcode",  # (5) Source code links
]

# -- AutoAPI configuration ---------------------------------------------------
autoapi_type = "python"
autoapi_dirs = ["../../open_autonlu"]
autoapi_root = "api"
autoapi_options = [
    "members",
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
    "inherited-members",  # Filtered conditionally in skip_member
]
autoapi_ignore = [
    "*/__pycache__/*",
    "*/.env*",
    # Excluded modules
    "*/auto_classes/constants.py",
    "*/data_quality/data_types.py",
    "*/data_quality/utils.py",
    "*/data_types.py",
    "*/llm_pipelines/processor.py",
    "*/methods/utils/*",
    "*/progress_reporter.py",
]
autoapi_python_class_content = "both"

# Classes that should show inherited members
SHOW_INHERITED_FOR = [
    "open_autonlu.auto_classes",
    "open_autonlu.methods.finetuner_ood",
    "open_autonlu.methods.set_fit_ood",
    "open_autonlu.methods.anc_set_fit_ood",
]

# Attributes to always skip (like logger instances)
SKIP_ATTRIBUTES = ["log", "logger", "logging"]


# (2) Hide private members, (3) Hide abstract methods
def skip_member(app, what, name, obj, skip, options):
    # Get short name (last part after dot)
    short_name = name.split(".")[-1]
    # Skip private members (but keep __init__ and other dunder if needed)
    if short_name.startswith("_") and not short_name.startswith("__"):
        return True
    # Skip logger attributes
    if short_name in SKIP_ATTRIBUTES:
        return True
    # Skip abstract methods
    if (
        hasattr(obj, "properties")
        and obj.properties
        and "abstractmethod" in obj.properties
    ):
        return True
    # Skip inherited members unless in allowed modules
    if hasattr(obj, "inherited") and obj.inherited:
        # Check if this belongs to a module that should show inherited
        should_show = any(name.startswith(prefix) for prefix in SHOW_INHERITED_FOR)
        if not should_show:
            return True
    # Skip attributes without docstrings (empty attributes)
    if what == "attribute" and hasattr(obj, "docstring") and not obj.docstring:
        return True
    return skip


def setup(app):
    app.connect("autoapi-skip-member", skip_member)


# -- Autodoc configuration ---------------------------------------------------
autodoc_member_order = "bysource"
autodoc_inherit_docstrings = True

templates_path = ["_templates"]
exclude_patterns = []
language = "en"

# -- Options for HTML output -------------------------------------------------
html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_css_files = ["custom.css"]



from __future__ import annotations

import logging
import re
from typing import Dict, Optional, Tuple

import pandas as pd

from .configs.llm_config import LlmClientConfig
from .model_endpoint import ModelEndpointWrapper
from .prompts import get_prompts

log = logging.getLogger(__name__)


def analyze_domain_and_labels(
    reference_data: pd.DataFrame,
    text_column: str = "text",
    label_column: str = "label",
    examples_per_label: int = 5,
    config_overrides: Optional[Dict] = None,
    language: str = "en",
) -> Tuple[str, Dict[str, str]]:
    """
    Analyzes domain and labels using prompt analyzer.

    Args:
        reference_data: DataFrame with data for analysis
        text_column: Name of text column
        label_column: Name of label column
        examples_per_label: Number of examples per label for analysis
        config_overrides: Optional client configuration overrides

    Returns:
        Tuple[str, Dict[str, str]]: (domain description, dictionary of label descriptions)
    """
    prompts = get_prompts(language)

    examples_by_label = []
    for label in reference_data[label_column].unique():
        label_data = reference_data[reference_data[label_column] == label]
        num_examples = min(examples_per_label, len(label_data))
        if num_examples == 0:
            log.warning("Label '%s' has no examples. Skipping.", label)
            continue
        sample_texts = label_data.sample(n=num_examples)[text_column].tolist()

        examples_by_label.append(f"{prompts.label_prefix}: {label}")
        for text in sample_texts:
            examples_by_label.append(f"- {text}")
        examples_by_label.append("")

    examples_text = "\n".join(examples_by_label)

    labels_list = [str(label) for label in reference_data[label_column].unique()]
    label_names = ", ".join(labels_list)

    if config_overrides:
        llm_kwargs = config_overrides.get("LlmClientConfig", {})
        client_config = LlmClientConfig(**llm_kwargs)
    else:
        client_config = LlmClientConfig()

    client_config.default_system_prompt = prompts.analyzer_system_prompt
    client = ModelEndpointWrapper(config=client_config)

    log.info("Starting domain and labels analysis...")
    prompt_text = prompts.analyze_domain(
        examples_by_label=examples_text, label_names=label_names
    )

    try:
        response = client(prompt_text, system_prompt=prompts.analyzer_system_prompt)
        log.info("Received analyzer response (%s characters)", len(response))
    except Exception as exc:  # pylint: disable=broad-except
        log.error("Error calling analyzer: %s", exc, exc_info=True)
        return "", {}

    # Ensure labels are strings for regex matching
    labels = [str(label) for label in reference_data[label_column].unique().tolist()]
    domain_desc, label_descriptions = _parse_analysis_response(
        response=response,
        labels=labels,
        default_label_desc_template=prompts.default_label_desc_template,
        domain_description_header=prompts.domain_description_header,
        label_descriptions_header=prompts.label_descriptions_header,
    )
    log.info("Extracted domain description (%s characters)", len(domain_desc))
    log.info("Extracted label descriptions: %s", len(label_descriptions))

    # which labels were not found
    missing_labels = [label for label in labels if label not in label_descriptions]
    if missing_labels:
        log.warning("Could not extract descriptions for labels: %s", missing_labels)
        labels_header = get_prompts(language).label_descriptions_header
        if labels_header in response:
            labels_start = response.find(labels_header)
            labels_section = response[labels_start : labels_start + 3000]
            log.warning(
                "%s (first 3000 chars):\n%s",
                labels_header,
                labels_section,
            )
        else:
            log.warning("%s section not found in response", labels_header)

    return domain_desc, label_descriptions


def _parse_analysis_response(
    response: str,
    labels: list[str],
    default_label_desc_template: str = (
        "Texts of category '{label}' correspond to the general stylistic and "
        "thematic features of this text corpus, while having "
        "characteristic content related to the topic {label}."
    ),
    domain_description_header: str = "DOMAIN DESCRIPTION:",
    label_descriptions_header: str = "LABEL DESCRIPTIONS:",
) -> Tuple[str, Dict[str, str]]:
    domain_desc = ""
    label_descriptions: Dict[str, str] = {}

    if response:
        domain_pattern = (
            re.escape(domain_description_header)
            + r"(.*?)(?="
            + re.escape(label_descriptions_header)
            + r"|$)"
        )
        domain_match = re.search(domain_pattern, response, re.DOTALL)
        if domain_match:
            domain_desc = domain_match.group(1).strip()

        labels_pattern = re.escape(label_descriptions_header) + r"(.*?)$"
        labels_section = re.search(labels_pattern, response, re.DOTALL)
        if labels_section:
            labels_text = labels_section.group(1).strip()

            for label in labels:
                pattern = rf"{re.escape(label)}:\s*(.*?)(?=\n\n[a-zA-Z0-9_]+:|$)"
                label_match = re.search(pattern, labels_text, re.DOTALL)

                if label_match:
                    desc = label_match.group(1).strip()
                    if desc:
                        label_descriptions[label] = desc
                        log.info(
                            "Extracted description for label '%s': %s characters",
                            label,
                            len(desc),
                        )
                    else:
                        log.warning(
                            "WARNING: Description for label '%s' is empty",
                            label,
                        )
                else:
                    pattern2 = rf"{re.escape(label)}:(.*?)(?=\n[a-zA-Z0-9_]+:|$)"
                    label_match2 = re.search(pattern2, labels_text, re.DOTALL)
                    if label_match2:
                        desc = label_match2.group(1).strip()
                        if desc:
                            label_descriptions[label] = desc
                            log.info(
                                "Extracted description for label '%s' (alternative method): %s characters",
                                label,
                                len(desc),
                            )
                        else:
                            log.warning(
                                "WARNING: Description for label '%s' is empty (alternative method)",
                                label,
                            )
                    else:
                        log.warning(
                            "WARNING: Failed to extract description for label '%s'",
                            label,
                        )

        if not label_descriptions:
            log.error("Critical error: Failed to extract any label descriptions!")
            log.error("Full model response:\n%s", response)

            for label in labels:
                lines_list = response.split("\n")
                for i, line in enumerate(lines_list):
                    if line.strip().startswith(f"{label}:"):
                        desc_lines = []
                        j = i + 1
                        while j < len(lines_list) and not any(
                            lines_list[j].strip().startswith(f"{other_label}:")
                            for other_label in labels
                        ):
                            if lines_list[j].strip():
                                desc_lines.append(lines_list[j])
                            j += 1

                        if desc_lines:
                            label_descriptions[label] = "\n".join(desc_lines).strip()
                            log.info(
                                "Extracted description for label '%s' (line-by-line method): %s characters",
                                label,
                                len(label_descriptions[label]),
                            )

        log.info("Total extracted label descriptions: %s", len(label_descriptions))
        if len(label_descriptions) < len(labels):
            log.warning(
                "WARNING: Only extracted %s descriptions out of %s labels!",
                len(label_descriptions),
                len(labels),
            )

        for label in labels:
            if label not in label_descriptions or not label_descriptions[label]:
                default_desc = default_label_desc_template.format(label=label)
                label_descriptions[label] = default_desc
                log.warning("Created default description for label '%s'", label)

    return domain_desc, label_descriptions



import re
import random
from typing import Any, Iterable, Iterator, List, Union, Optional, Callable
import mdpd
import numpy as np
import outlines
import pandas as pd
import tqdm
from .model_endpoint import ModelEndpointWrapper
from openai import APIConnectionError, BadRequestError
import threading
import logging

log = logging.getLogger(__name__)


class WrongFormatError(Exception):
    pass


class Processor:
    def __init__(
        self,
        prompt: outlines.prompts.Prompt,
        client: ModelEndpointWrapper,
        seed: int,
        prompt_takes_batches: bool,
        num_retries: int = 3,
        batch_size: int = None,
    ):
        self.prompt = prompt
        self.batch_size = batch_size
        self.client = client
        self.seed = seed
        self.prompt_takes_batches = prompt_takes_batches
        self.num_retries = num_retries

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        """
        Process texts either in batches or individually, handling retries for wrongly formatted responses.

        Parameters:
        *args: Variable length argument list.
        **kwds: Arbitrary keyword arguments.

        Returns:
        Any: The aggregated or processed responses.
        """
        if self.prompt_takes_batches:
            return self._process_batches(**kwds)
        else:
            return self._process_individual(**kwds)

    def _api_call(
        self,
        prompt: Union[str, outlines.prompt],
        postprocessor: Optional[Callable] = lambda x: x,
        stop_flag: Optional[threading.Event] = None,
    ) -> Any:
        """
        Processes a request to the model's API and handles any errors that occur.

        Args:
            prompt (str): The instruction prompt for the generative model.
            postprocessor (Callable): Function to process response from generative model.
            stop_flag (threading.Event): A flag used to stop API calls running in different threads.

        Returns:
            Any: Post-processed response from the generative model or None in case of failure.
        """
        for trial in range(self.num_retries):
            if stop_flag is not None and stop_flag.is_set():
                return None
            try:
                response = self.client(prompt)
                response = postprocessor(response)
                return response
            except WrongFormatError:
                log.info(
                    f"Trial {trial + 1}. LLM has output wrongly formatted text, retrying."
                )
            except APIConnectionError as e:
                raise e
            except BadRequestError as e:
                log.error("Bad request. Skipping this request.", exc_info=e)
                return None
            except Exception as e:
                log.error(
                    f"{e}\nTrial {trial + 1}. An exception occurred while calling LLM, retrying.",
                    exc_info=e,
                )
        else:
            log.info("Out of retries, skipping request.")
            return None

    def _process_batches(self, **kwds: Any) -> Any:
        """
        Process texts in batches, handling retries for wrongly formatted responses.

        Parameters:
        **kwds: Arbitrary keyword arguments.

        Returns:
        Any: The aggregated batches.
        """
        postprocessed_batches = []
        texts = kwds.pop("texts")
        batch_size = kwds.pop("batch_size", self.batch_size)
        stop_event = kwds.pop("stop_event", None)
        example_texts = kwds.pop("example_texts", None)
        examples_per_batch = kwds.pop("examples_per_batch", None)

        assert (
            batch_size is not None
        ), "You have to specify batch size either in call args, or in constructor args"

        if example_texts is not None and examples_per_batch is not None:
            data_size = kwds.get("data_size", batch_size)
            num_batches = (data_size + batch_size - 1) // batch_size

            rng = random.Random(self.seed)

            for batch_idx in range(num_batches):
                # Calculate how many texts to generate in this batch
                remaining = data_size - (batch_idx * batch_size)
                current_batch_size = min(batch_size, remaining)
                if current_batch_size <= 0:
                    break

                num_examples = min(examples_per_batch, len(example_texts))
                batch_examples = rng.sample(example_texts, num_examples)
                rng.shuffle(batch_examples)

                # Pass current_batch_size instead of total data_size
                batch_kwds = kwds.copy()
                batch_kwds["data_size"] = current_batch_size

                response = self._api_call(
                    self.prompt(texts=batch_examples, **batch_kwds),
                    postprocessor=self._postprocess_one_batch,
                    stop_flag=stop_event,
                )
                if response is not None:
                    postprocessed_batches.append(response)
        else:
            for batch in self._window(texts, batch_size, self.seed):
                response = self._api_call(
                    self.prompt(texts=batch, **kwds),
                    postprocessor=self._postprocess_one_batch,
                    stop_flag=stop_event,
                )
                if response is not None:
                    postprocessed_batches.append(response)

        if len(postprocessed_batches) == 0:
            return None
        else:
            return self._aggregate_batches(postprocessed_batches)

    def _process_individual(self, **kwds: Any) -> Any:
        """
        Process texts individually.

        Parameters:
        **kwds: Arbitrary keyword arguments.

        Returns:
        Any: The aggregated individual responses.
        """
        postprocessed_responses = []
        texts = kwds.pop("texts")
        stop_event = kwds.pop("stop_event", None)

        for text in tqdm.tqdm(texts, desc="Processing texts"):
            response = self._api_call(
                self.prompt(text=text, **kwds),
                postprocessor=self._postprocess_one,
                stop_flag=stop_event,
            )
            if response is not None:
                postprocessed_responses.append(response)
        if len(postprocessed_responses) == 0:
            return None
        else:
            return self._aggregate_single_responses(postprocessed_responses, texts)

    @staticmethod
    def _window(
        texts: Iterable[str], subset_size: int, seed: int
    ) -> Iterator[List[str]]:
        """
        Divides an iterable of text strings into subsets of a specified size.

        The method uses a random number generator to ensure the subsets are
        distributed randomly. It yields each subset as a list of strings.

        Parameters:
        -----------
        texts : Iterable[str]
            An iterable of text strings that need to be divided into subsets.
        subset_size : int
            The size of each subset. This value must be less than or equal to
            the number of text samples provided.
        seed : int
            A seed value for the random number generator to ensure reproducibility
            of the subset distribution.

        Returns:
        --------
        Iterator[List[str]]
            An iterator that yields lists of strings, where each list represents
            a subset of the input texts.

        Raises:
        -------
        AssertionError
            If the `subset_size` is greater than the number of text samples provided.
        """
        df = pd.DataFrame({"text": texts})
        # TODO: process this assertion error in self._process_batches() and self._process_individual()
        assert (
            subset_size <= df.shape[0]
        ), f"Choose a smaller window, your data is only {df.shape[0]} samples"
        rng = np.random.default_rng(seed=seed)
        n_samples = df.shape[0] // subset_size
        df["subset"] = rng.choice(n_samples, df.shape[0])
        for _, group in df.groupby("subset"):
            yield group.text.tolist()

    @staticmethod
    def _postprocess_one_batch(text_response: str):
        # response should contain only a table
        raw_table = re.search(r"\|.*\|", text_response, re.S)
        if raw_table:
            return mdpd.from_md(raw_table.group(0))
        else:
            raise WrongFormatError(text_response)

    def _aggregate_batches(self, batches: List[pd.DataFrame]) -> pd.DataFrame:
        raise NotImplementedError()

    def _aggregate_single_responses(
        self, responses: List[str], texts: List[str]
    ) -> pd.DataFrame:
        raise NotImplementedError()

    def _postprocess_one(self, response: List[str]) -> Any:
        raise NotImplementedError()



import enum
from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional, Union

import numpy as np
from torch.utils.data import Dataset


@dataclass
class Entity:
    """Represents a named entity extracted from text.

    Attributes:
        label: Entity type/class label (e.g., "PERSON", "ORG").
        text: The actual text span of the entity.
        start: Character offset where the entity starts in the source text.
        end: Character offset where the entity ends in the source text.
        score: Confidence score for the entity prediction.
    """

    label: str
    text: str
    start: int
    end: int
    score: float


class NERFormat(enum.Enum):
    """Format for Named Entity Recognition annotations.

    Attributes:
        OFFSETS: Character offset format (start, end positions).
        BIO: BIO tagging format (Beginning-Inside-Outside).
    """

    OFFSETS = "offsets"
    BIO = "bio"


class SaveFormat(enum.Enum):
    """Model serialization format.

    Attributes:
        TORCH: PyTorch native format (.pt/.pth files).
        ONNX: ONNX format.
    """

    TORCH = "torch"
    ONNX = "onnx"


class OodMethod(enum.Enum):
    """Out-of-Distribution detection method.

    Attributes:
        NONE: No OOD detection.
        LOGIT: Logit-based OOD (adds an out-of-scope class during training).
        AUTO: Automatically select method based on the training approach.
        MARGINAL_MAHALANOBIS_OOD: Mahalanobis distance-based detection.
        MSP_OOD: Maximum Softmax Probability - uses classifier confidence
            as OOD score.
    """

    NONE = "none"
    LOGIT = "logit"
    AUTO = "auto"
    MARGINAL_MAHALANOBIS_OOD = "MarginalMahalanobisTrainer"
    MSP_OOD = "MaxSoftmaxProbabilityTrainer"


@dataclass
class TokenClassifierOutput:
    """Output from a token classification (NER) model.

    Attributes:
        text: The original input text.
        labels: Extracted entities as Entity objects (OFFSETS format) or
            BIO tags as strings (BIO format).
    """

    text: str
    labels: Union[List[Entity], List[str]]


@dataclass
class TextLabel:
    """A single classification hypothesis with confidence score.

    Attributes:
        label: The predicted class label.
        score: Confidence score for this prediction.
    """

    label: str
    score: float


@dataclass
class TextClassifierOutput:
    """Output from a text classification model.

    Contains the input text and ranked list of classification hypotheses.

    Attributes:
        text: The original input text.
        hypotheses: List of TextLabel predictions, sorted by score descending.
    """

    text: str
    hypotheses: List[TextLabel]

    @property
    def most_probable(self) -> TextLabel:
        """Returns the highest-scoring classification hypothesis."""
        return self.hypotheses[0]


class SimplePyTorchDataset(Dataset):
    """Simple PyTorch Dataset wrapper for dictionary-based data.

    Wraps a dictionary of column arrays into a PyTorch Dataset interface.
    Each item returns a dictionary with values at the given index from
    each column.

    Attributes:
        data: Dictionary mapping column names to iterable values.

    Example:
        >>> data = {"text": ["hello", "world"], "label": [0, 1]}
        >>> dataset = SimplePyTorchDataset(data)
        >>> dataset[0]  # {"text": "hello", "label": 0}
    """

    def __init__(self, columns_dict: Dict[str, Iterable]):
        self.data = columns_dict

    def __getitem__(self, index) -> Dict[str, Any]:
        item = {}
        for key, value in self.data.items():
            item[key] = value[index]
        return item

    def __len__(self):
        first_key = list(self.data.keys())[0]
        return len(self.data[first_key])


@dataclass
class ClassificationEvaluationResult:
    """Evaluation metrics for a classification model.

    Attributes:
        confusion_matrix: NxN confusion matrix as numpy array.
        classification_report: Per-class metrics (precision, recall, f1, support).
        f1: Macro-averaged F1 score.
    """

    confusion_matrix: np.ndarray
    classification_report: Dict[str, Dict[str, float]]
    f1: float


@dataclass
class RetrievalEvaluationResult:
    recall_at_1: float


@dataclass
class TrainingArtifactInfo:
    """Information about a trained model artifact.

    Captures hyperparameters, labels, and evaluation metrics from training.
    Returned by training methods to provide metadata about the trained model.

    Attributes:
        hyperparameters: Dictionary of hyperparameters used for training.
        labels: List of class labels the model was trained on.
        dev_metrics: Evaluation results on development/validation set.
        test_metrics: Evaluation results on test set (all samples).
        test_metrics_inscope: Evaluation results on in-scope test samples only
            (excludes OOD samples).
    """

    hyperparameters: Dict[str, Any]
    labels: List[str]
    dev_metrics: Optional[ClassificationEvaluationResult] = None
    test_metrics: Optional[
        Union[ClassificationEvaluationResult, RetrievalEvaluationResult]
    ] = None
    test_metrics_inscope: Optional[ClassificationEvaluationResult] = None



DEFAULT_MODEL_BY_LANGUAGE: dict[str, str] = {
    "ru": "ai-forever/ruBert-base",
    "en": "bert-base-uncased",
}

DEFAULT_TEMPLATE_BY_LANGUAGE: dict[str, str] = {
    "ru": "Пользователь просит бота выполнить запрос с помощью навыка: ",
    "en": "User asks the bot to perform a request using the skill: ",
}
MARGIN = 0.25
NUM_ITERATION = 20
MAX_SEQ_LENGTH = 512
ADD_NORMALIZATION_LAYER = False
BATCH_SIZE = 64
EPOCHS = 1
LR = 1e-4
SEED = 1223
OOS_LABEL = "outOfScope"



from __future__ import annotations

import glob
import json
import logging
import math
import os
import warnings
from collections import Counter
from collections.abc import Iterable
from typing import Any, Dict, List, Union

from datasets import Dataset, DatasetDict

from ..constants import DQ_TRAIN
from ..methods.constants import OOS_LABEL
from ..progress_reporter import ProgressReporter
from .data_types import DatasetEvaluatorOutput, TaskType
from .dynamic_finetuner import DynamicTuner
from .sample_eval import AbstractSampleEvalMethod

log = logging.getLogger(__name__)
warnings.filterwarnings("ignore")


class DatasetEvaluator:
    """Evaluates dataset quality and identifies problematic samples.

    Runs multiple sample evaluation methods (Retag, Uncertainty, VInfo,
    LabelAggregation) to identify samples that may be mislabeled,
    ambiguous, or otherwise problematic for training.
    """

    def __init__(
        self,
        task_type: Union[str, TaskType],
        tuner: DynamicTuner,
        sample_evaluators: Iterable[AbstractSampleEvalMethod] | None,
        workdir: str = "./data_quality_output",
        save_results: bool = True,
        aggregate: str = "once",
    ):
        """Initialize the dataset evaluator.

        Args:
            task_type: Type of task ("multiclass_classification" or
                "token_classification").
            tuner: DynamicTuner instance for model-based evaluation.
            sample_evaluators: Iterable of evaluation methods to apply.
            workdir: Directory for saving evaluation results.
            save_results: Whether to save results to disk.
            aggregate: Aggregation strategy for multiple evaluators:
                - 'once': Flag if any method marks as problematic
                - 'all': Flag only if all methods agree
                - 'vote': Flag if majority of methods agree
        """
        self.task_type = (
            TaskType(task_type) if not isinstance(task_type, TaskType) else task_type
        )
        self.sample_evaluators: Iterable[AbstractSampleEvalMethod] = (
            self._validate_evaluators(sample_evaluators)
        )
        self.workdir = workdir
        os.makedirs(self.workdir, exist_ok=True)
        self.save_results = save_results
        self.aggregate = aggregate
        self.tuner = tuner
        self.id2label = None
        self.aggregated_report_filename = "aggregated_report.csv"

    # TODO: need more advanced validation, pydantic fields or smth like this
    def _validate_evaluators(
        self, sample_evaluators: Any
    ) -> Iterable[AbstractSampleEvalMethod]:
        """Validate and transform format to list of methods, or empty list

        Args:
            sample_evaluators (Iterable[AbstractSampleEvalMethod] | None): sample_evaluators

        Returns:
            list: sample_evaluators
        """
        if sample_evaluators is None:
            return []

        if isinstance(sample_evaluators, Iterable):
            for evaluator in sample_evaluators:
                if self.task_type not in evaluator.supported_tasks:
                    raise ValueError(
                        f"{evaluator.name} is not supported for {self.task_type} task"
                    )

            return sample_evaluators

        raise TypeError(
            f"sample_evaluators should be None, Iterable, not {type(sample_evaluators)}"
        )

    def filter_data(
        self, dataset: DatasetDict | Dataset, progress_reporter: ProgressReporter
    ) -> DatasetEvaluatorOutput:
        """The main method to filter dataset

        Args:
            dataset (DatasetDict | Dataset): dataset

        Returns:
            DatasetEvaluatorOutput
        """
        assert progress_reporter.stages is not None

        # no ood samples for data quality
        if isinstance(dataset, DatasetDict):
            filtered_splits = DatasetDict()
            for split_name, split_ds in dataset.items():
                if "label" in split_ds.column_names:
                    filtered_splits[split_name] = split_ds.filter(
                        lambda example: example["label"] != OOS_LABEL
                    )
                else:
                    filtered_splits[split_name] = split_ds
            dataset = filtered_splits
        elif isinstance(dataset, Dataset):
            if "label" in dataset.column_names:
                dataset = dataset.filter(lambda example: example["label"] != OOS_LABEL)

        # all evaluators, methods need to identify sample
        dataset = dataset.map(lambda _, idx: {"id": idx}, with_indices=True)

        if self._need_tuned_model():
            progress_reporter.stages = [DQ_TRAIN, *progress_reporter.stages]
            self._tune_model(dataset, progress_reporter)

        counter = Counter()
        method_candidates = {}
        for evaluator in self.sample_evaluators:
            progress_reporter(evaluator.name, 0.0)
            log.info(f"{evaluator.name:=^20}")
            tuner = self.tuner if evaluator.need_tuner_instance else None
            # TODO: передавать только нужные параметры
            indices_to_remove: List[int] = evaluator.candidates_to_remove(
                dataset,
                workdir=self.workdir,
                tuner=tuner,
                id2label=self.id2label,
                task_type=self.task_type,
            )
            method_candidates[evaluator.name] = indices_to_remove
            log.info(f"Candidates: {len(indices_to_remove)}")
            counter.update(indices_to_remove)
            progress_reporter(evaluator.name, 1.0)

        log.info(f"{'method to aggregate: ' + self.aggregate:=^10}")
        if self.aggregate == "all":
            counter = dict(
                filter(lambda x: x[1] == len(self.sample_evaluators), counter.items())
            )
        if self.aggregate == "vote":
            counter = dict(
                filter(
                    lambda x: x[1] >= math.ceil(len(self.sample_evaluators) / 2),
                    counter.items(),
                )
            )

        aggregate_indices = counter.keys()

        log.info(f"Final length of indices to remove: {len(aggregate_indices)}")

        dataset = dataset.filter(
            lambda _, idx: idx not in aggregate_indices, with_indices=True
        )
        dataset = dataset.remove_columns("id")
        if self.save_results:
            self._save(aggregate_indices, method_candidates)
            if self.aggregated_report_filename is None:
                aggregated_report_path = None
            else:
                aggregated_report_path = os.path.join(
                    self.workdir, self.aggregated_report_filename
                )
            return DatasetEvaluatorOutput(
                splits_filtered=dataset,
                indices_to_remove=aggregate_indices,
                aggregated_report_path=aggregated_report_path,
                reports_by_method_folder=self.workdir,
            )

        return DatasetEvaluatorOutput(
            splits_filtered=dataset,
            indices_to_remove=aggregate_indices,
            aggregated_report_path=None,
            reports_by_method_folder=None,
        )

    def _save(
        self, aggregate_indices: Iterable[int], method_candidates: Dict[str, List[int]]
    ):
        files_to_merge = None
        common_columns = ["id_sample", "text", "label"]
        if self.task_type == TaskType.NER:
            common_columns.append("id_token")

        for evaluator in self.sample_evaluators:
            evaluator.save(self.workdir)
            if evaluator.prepared_data.shape[0] > 0:
                evaluator.prepared_data[f"method_{evaluator.name}"] = evaluator.name
                if files_to_merge is None:
                    files_to_merge = evaluator.prepared_data
                    files_to_merge.columns = [
                        col if col in common_columns else col + "_" + evaluator.name
                        for col in files_to_merge.columns
                    ]
                else:
                    evaluator.prepared_data.columns = [
                        col if col in common_columns else col + "_" + evaluator.name
                        for col in evaluator.prepared_data.columns
                    ]
                    files_to_merge = files_to_merge.merge(
                        evaluator.prepared_data,
                        how="outer",
                        on=common_columns,
                        suffixes=("", "_" + evaluator.name),
                    )

        if files_to_merge is not None:
            method_columns = [
                c for c in files_to_merge.columns if c.startswith("method")
            ]
            files_to_merge[method_columns] = files_to_merge[method_columns].fillna("")
            files_to_merge["methods"] = files_to_merge[method_columns].agg(
                " ".join, axis=1
            )
            files_to_merge["methods"] = (
                files_to_merge["methods"].replace(r"\s+", " ", regex=True).str.strip()
            )
            files_to_merge = files_to_merge.drop(method_columns, axis=1)

            files_to_merge.to_csv(
                os.path.join(self.workdir, self.aggregated_report_filename), index=False
            )
            with open(
                os.path.join(self.workdir, "candidates_to_remove.json"), "w"
            ) as file:
                json.dump(method_candidates, file)
            with open(
                os.path.join(self.workdir, "aggregated_candidates.json"), "w"
            ) as file:
                json.dump(list(aggregate_indices), file)
        else:
            self.aggregated_report_filename = None

    def _need_tuned_model(self) -> bool:
        return any([eval.need_tuned_model for eval in self.sample_evaluators])

    def _tune_model(
        self, dataset: DatasetDict | Dataset, progress_reporter: ProgressReporter
    ):
        if progress_reporter is None:
            progress_reporter = ProgressReporter()

        filepaths = glob.glob(os.path.join(f"{self.workdir}", "**"), recursive=True)
        for filepath in filepaths:
            if os.path.isfile(filepath):
                os.remove(filepath)
                log.info(f"removed previous calculations: {filepath}")

        self.tuner.training_args.output_dir = self.workdir
        self.tuner.train(dataset["train"], progress_reporter)
        # token classification naming
        if hasattr(self.tuner, "label_names"):
            self.id2label = {k: v for k, v in enumerate(self.tuner.label_names)}
        # sequence classification naming
        if hasattr(self.tuner, "id2label"):
            self.id2label = self.tuner.id2label
